from __future__ import absolute_import, division, print_function, unicode_literals

from unittest.case import TestCase
from uuid import uuid4

from eventsourcing.domain.model.collection import Collection, register_new_collection
from eventsourcing.domain.model.entity import EntityIsDiscarded
from eventsourcing.domain.model.events import assert_event_handlers_empty, subscribe, unsubscribe
from eventsourcing.exceptions import RepositoryKeyError
from eventsourcing.infrastructure.event_sourced_repos.collection_repo import CollectionRepository
from eventsourcing.tests.sequenced_item_tests.base import WithPersistencePolicy
from eventsourcing.tests.sequenced_item_tests.test_sqlalchemy_active_record_strategy import \
    WithSQLAlchemyActiveRecordStrategies


class TestCollection(TestCase):

    def setUp(self):
        assert_event_handlers_empty()
        self.published_events = []
        self.subscription = (lambda x: True, lambda x: self.published_events.append(x))
        subscribe(*self.subscription)

    def tearDown(self):
        unsubscribe(*self.subscription)
        assert_event_handlers_empty()

    def test(self):
        # Register a new collection entity.
        collection_id = 'collection1'
        collection = register_new_collection(collection_id=collection_id)

        # Check collection ID is set.
        self.assertEqual(collection.id, collection_id)

        # Declare items.
        item1 = 'item1'
        item2 = 'item2'

        # Check the collection is empty.
        self.assertNotIn(item1, collection)
        self.assertNotIn(item2, collection)
        self.assertEqual(len(collection._items), 0)

        # Check there has been one Collection.Created event.
        self.assertEqual(len(self.published_events), 1)
        last_event = self.published_events[-1]
        self.assertIsInstance(last_event, Collection.Created)
        self.assertEqual(last_event.entity_id, collection_id)

        # Add item to collection.
        collection.add_item(item1)

        # Check the collection has item1 only.
        self.assertIn(item1, collection)
        self.assertNotIn(item2, collection)
        self.assertEqual(len(collection._items), 1)

        # Check there has been one Collection.ItemAdded event.
        self.assertEqual(len(self.published_events), 2)
        last_event = self.published_events[-1]
        self.assertIsInstance(last_event, Collection.ItemAdded)
        self.assertEqual(last_event.entity_id, collection_id)
        self.assertEqual(last_event.item, item1)

        # Add another item.
        collection.add_item(item2)

        # Check the collection has item1 and item2.
        self.assertIn(item1, collection)
        self.assertIn(item2, collection)
        self.assertEqual(len(collection._items), 2)

        # Check there has been another Collection.ItemAdded event.
        self.assertEqual(len(self.published_events), 3)
        last_event = self.published_events[-1]
        self.assertIsInstance(last_event, Collection.ItemAdded)
        self.assertEqual(last_event.entity_id, collection_id)
        self.assertEqual(last_event.item, item2)

        # Remove item1 from the collection.
        collection.remove_item(item1)

        # Check the collection has item2 only.
        self.assertNotIn(item1, collection)
        self.assertIn(item2, collection)
        self.assertEqual(len(collection._items), 1)

        # Check there has been a Collection.ItemRemoved event.
        self.assertEqual(len(self.published_events), 4)
        last_event = self.published_events[-1]
        self.assertIsInstance(last_event, Collection.ItemRemoved)
        self.assertEqual(last_event.entity_id, collection_id)
        self.assertEqual(last_event.item, item1)

        # Discard the collection.
        collection.discard()

        # Check there has been a Collection.Discarded event.
        self.assertEqual(len(self.published_events), 5)
        last_event = self.published_events[-1]
        self.assertIsInstance(last_event, Collection.Discarded)
        self.assertEqual(last_event.entity_id, collection_id)

        self.assertRaises(EntityIsDiscarded, getattr, collection, 'items')


class TestCollectionRepo(WithSQLAlchemyActiveRecordStrategies, WithPersistencePolicy):

    def test(self):

        repo = CollectionRepository(event_store=self.versioned_entity_event_store)

        # Check unknown collections are not found in the repo.
        with self.assertRaises(RepositoryKeyError):
            _ = repo[uuid4()]

        # Register a new collection.
        collection_id = register_new_collection().id

        # Check the collection is in the repo.
        collection = repo[collection_id]
        self.assertIsInstance(collection, Collection)
        self.assertEqual(collection.id, collection_id)
        # Check the collection has zero items.
        self.assertEqual(len(collection.items), 0)

        # Add item.
        item1 = 'item1'
        collection.add_item(item1)

        # Check the collection is in the repo.
        collection = repo[collection_id]
        self.assertIsInstance(collection, Collection)
        self.assertEqual(collection.id, collection_id)
        # Check the collection has one item.
        self.assertEqual(len(collection.items), 1)

        # Remove item.
        collection.remove_item(item1)

        # Check the collection is in the repo.
        collection = repo[collection_id]
        self.assertIsInstance(collection, Collection)
        self.assertEqual(collection.id, collection_id)
        # Check the collection has zero items.
        self.assertEqual(len(collection.items), 0)

        # Discard the collection.
        collection.discard()

        # Check the collection is not in the repo.
        with self.assertRaises(RepositoryKeyError):
            _ = repo[collection.id]



import datetime
from decimal import Decimal
from time import sleep, time
from unittest.case import TestCase
from uuid import uuid4

from eventsourcing.domain.model.events import DomainEvent, TimestampedEntityEvent, VersionedEntityEvent, \
    topic_from_domain_class
from eventsourcing.infrastructure.sequenceditem import SequencedItem
from eventsourcing.infrastructure.sequenceditemmapper import SequencedItemMapper


class Event1(VersionedEntityEvent):
    pass


class Event2(TimestampedEntityEvent):
    pass


class Event3(DomainEvent):
    pass


class ValueObject1(object):
    def __init__(self, value):
        self.value = value

    def __eq__(self, other):
        return self.__dict__ == other.__dict__


class TestSequencedItemMapper(TestCase):
    def test_with_versioned_entity_event(self):
        # Setup the mapper, and create an event.
        mapper = SequencedItemMapper(
            sequenced_item_class=SequencedItem,
            event_sequence_id_attr='entity_id',
            event_position_attr='entity_version'
        )
        entity_id1 = uuid4()
        event1 = Event1(entity_id=entity_id1, entity_version=101)

        # Check to_sequenced_item() method results in a sequenced item.
        sequenced_item = mapper.to_sequenced_item(event1)
        self.assertIsInstance(sequenced_item, SequencedItem)
        self.assertEqual(sequenced_item.position, 101)
        self.assertEqual(sequenced_item.sequence_id, entity_id1)
        self.assertEqual(sequenced_item.topic, topic_from_domain_class(Event1))
        self.assertTrue(sequenced_item.data)

        # Use the returned values to create a new sequenced item.
        sequenced_item_copy = SequencedItem(
            sequence_id=sequenced_item.sequence_id,
            position=sequenced_item.position,
            topic=sequenced_item.topic,
            data=sequenced_item.data,
        )

        # Check from_sequenced_item() returns an event.
        domain_event = mapper.from_sequenced_item(sequenced_item_copy)
        self.assertIsInstance(domain_event, Event1)
        self.assertEqual(domain_event.entity_id, event1.entity_id)
        self.assertEqual(domain_event.entity_version, event1.entity_version)

    def test_with_timestamped_entity_event(self):
        # Setup the mapper, and create an event.
        mapper = SequencedItemMapper(
            sequenced_item_class=SequencedItem,
            event_sequence_id_attr='entity_id',
            event_position_attr='timestamp'
        )
        before = time()
        sleep(0.000001)  # Avoid test failing due to timestamp having limited precision.
        event2 = Event2(entity_id='entity2')
        sleep(0.000001)  # Avoid test failing due to timestamp having limited precision.
        after = time()

        # Check to_sequenced_item() method results in a sequenced item.
        sequenced_item = mapper.to_sequenced_item(event2)
        self.assertIsInstance(sequenced_item, SequencedItem)
        self.assertGreater(sequenced_item.position, before)
        self.assertLess(sequenced_item.position, after)
        self.assertEqual(sequenced_item.sequence_id, 'entity2')
        self.assertEqual(sequenced_item.topic, topic_from_domain_class(Event2))
        self.assertTrue(sequenced_item.data)

        # Use the returned values to create a new sequenced item.
        sequenced_item_copy = SequencedItem(
            sequence_id=sequenced_item.sequence_id,
            position=sequenced_item.position,
            topic=sequenced_item.topic,
            data=sequenced_item.data,
        )

        # Check from_sequenced_item() returns an event.
        domain_event = mapper.from_sequenced_item(sequenced_item_copy)
        self.assertIsInstance(domain_event, Event2)
        self.assertEqual(domain_event.entity_id, event2.entity_id)
        self.assertEqual(domain_event.timestamp, event2.timestamp)

    def test_with_different_types_of_event_attributes(self):
        # Setup the mapper, and create an event.
        mapper = SequencedItemMapper(
            sequenced_item_class=SequencedItem,
            event_sequence_id_attr='entity_id',
            event_position_attr='a'
        )

        # Create an event with dates and datetimes.
        event3 = Event3(
            entity_id='entity3',
            entity_version=303,
            a=datetime.datetime(2017, 3, 22, 9, 12, 14),
            b=datetime.date(2017, 3, 22),
            c=uuid4(),
            # d=Decimal(1.1),
            e=ValueObject1('value1'),
        )

        # Check to_sequenced_item() method results in a sequenced item.
        sequenced_item = mapper.to_sequenced_item(event3)

        # Use the returned values to create a new sequenced item.
        sequenced_item_copy = SequencedItem(
            sequence_id=sequenced_item.sequence_id,
            position=sequenced_item.position,
            topic=sequenced_item.topic,
            data=sequenced_item.data,
        )

        # Check from_sequenced_item() returns an event.
        domain_event = mapper.from_sequenced_item(sequenced_item_copy)
        self.assertIsInstance(domain_event, Event3)
        self.assertEqual(domain_event.entity_id, event3.entity_id)
        self.assertEqual(domain_event.a, event3.a)
        self.assertEqual(domain_event.b, event3.b)
        self.assertEqual(domain_event.c, event3.c)
        # self.assertEqual(domain_event.d, event3.d)
        self.assertEqual(domain_event.e, event3.e)

    def test_errors(self):
        # Setup the mapper, and create an event.
        mapper = SequencedItemMapper(
            sequenced_item_class=SequencedItem,
            event_sequence_id_attr='entity_id',
            event_position_attr='a'
        )

        # Create an event with dates and datetimes.
        event3 = Event3(
            entity_id='entity3',
            entity_version=303,
            a=Decimal(1.0),
        )

        # Check to_sequenced_item() method results in a sequenced item.
        with self.assertRaises(TypeError):
            mapper.to_sequenced_item(event3)



import uuid

from eventsourcing.domain.model.entity import AbstractEntityRepository, Aggregate, AttributeChanged, Created, \
    Discarded, attribute, entity_mutator, singledispatch
from eventsourcing.domain.model.events import AggregateEvent, publish


class Example(Aggregate):
    """
    An example event sourced domain model entity.
    """

    class Created(Created):
        pass

    class AttributeChanged(AttributeChanged):
        pass

    class Discarded(Discarded):
        pass

    class Heartbeat(AggregateEvent):
        pass

    def __init__(self, a, b, **kwargs):
        super(Example, self).__init__(**kwargs)
        self._a = a
        self._b = b
        self._count_heartbeats = 0

    @attribute
    def a(self):
        """An example attribute."""

    @attribute
    def b(self):
        """Another example attribute."""

    def beat_heart(self, number_of_beats=1):
        self._assert_not_discarded()
        events = []
        while number_of_beats > 0:
            event = self.Heartbeat(entity_id=self._id, entity_version=self._version)
            events.append(event)
            self._apply(event)
            number_of_beats -= 1
        publish(events)

    def count_heartbeats(self):
        return self._count_heartbeats

    @staticmethod
    def _mutator(event, initial):
        return example_mutator(event, initial)


@singledispatch
def example_mutator(event, initial):
    return entity_mutator(event, initial)


@example_mutator.register(Example.Heartbeat)
def heartbeat_mutator(event, self):
    self._validate_originator(event)
    assert isinstance(self, Example), self
    self._count_heartbeats += 1
    self._increment_version()
    return self


class AbstractExampleRepository(AbstractEntityRepository):
    pass


def register_new_example(a, b):
    """
    Factory method for example entities.

    :rtype: Example
    """
    entity_id = uuid.uuid4()
    event = Example.Created(entity_id=entity_id, a=a, b=b)
    entity = Example.mutate(event=event)
    publish(event=event)
    return entity



from functools import reduce

from eventsourcing.tests.example_application_tests.base import WithExampleApplication
from eventsourcing.tests.sequenced_item_tests.test_sqlalchemy_active_record_strategy import \
    WithSQLAlchemyActiveRecordStrategies

# This tests using all the domain events in the application to project
# something other than an entity. It's not really customization.

# Todo: Support stopping and resuming when iterating over all events.

class TestGetAllEventFromSQLAlchemy(WithSQLAlchemyActiveRecordStrategies, WithExampleApplication):

    def test(self):
        with self.construct_application() as app:

            # Create three domain entities.
            entity1 = app.register_new_example('a1', 'b1')
            entity2 = app.register_new_example('a2', 'b2')
            entity3 = app.register_new_example('a3', 'b3')

            # Get all the domain events
            es = app.version_entity_event_store
            domain_events = es.all_domain_events()

            # Project the events into a set of entity IDs.
            def mutate(state, event):
                assert isinstance(state, set)
                state.add(event.entity_id)
                return state

            all_entity_ids = reduce(mutate, domain_events, set())

            # Check we got all the entity IDs.
            self.assertEqual(all_entity_ids, {entity1.id, entity2.id, entity3.id})

            # Todo: With integer sequenced items, to avoid keeping track of
            # a set of IDs, it would be possible to filter on position==0,
            # and either add that to the database query or put that in a
            # generator that yields all IDs of items with position==0 (rather
            # than yielding items that are not in a set, or keeping adding to
            # a set the same ID once for each version).



# from eventsourcing.domain.model.notificationlog import NotificationLog
# from eventsourcing.infrastructure.event_sourced_repos.timebucketedlog_repo import TimebucketedlogRepo
# from eventsourcing.infrastructure.event_sourced_repos.notificationlog_repo import NotificationLogRepo
# from eventsourcing.infrastructure.event_sourced_repos.sequence import SequenceRepo
# from eventsourcing.infrastructure.notification_log import NotificationLogReader, append_item_to_notification_log
# from eventsourcing.tests.sequenced_item_tests.test_cassandra_active_record_strategy import \
#     WithCassandraActiveRecordStrategies
# from eventsourcing.tests.sequenced_item_tests.test_sqlalchemy_active_record_strategy import \
#     WithSQLAlchemyActiveRecordStrategies
# from eventsourcing.tests.sequenced_item_tests.base import WithPersistencePolicy
#
#
# # Todo: Read from subsequent sequences if slice.stop goes beyond the end of the sequence.
#
#
# class NotificationLogTestCase(WithPersistencePolicy):
#
#     def test_entity_lifecycle(self):
#         notification_log_repo = NotificationLogRepo(self.timestamped_entity_event_store)
#
#         notification_log = notification_log_repo.get_or_create(
#             log_name='log1',
#             sequence_size=2,
#         )
#
#         self.assertIsInstance(notification_log, NotificationLog)
#
#         item1 = 'item1'
#
#         log_repo = TimebucketedlogRepo(self.timestamped_entity_event_store)
#         sequence_repo = SequenceRepo(event_store=self.versioned_entity_event_store)
#
#         append_item_to_notification_log(notification_log, item1, sequence_repo, log_repo, self.event_store)
#
#         notification_log_reader = NotificationLogReader(
#             notification_log=notification_log,
#             sequence_repo=sequence_repo,
#             log_repo=log_repo,
#             event_store=self.event_store,
#         )
#         self.assertEqual(notification_log_reader[0], item1)
#
#         item2 = 'item2'
#         append_item_to_notification_log(notification_log, item2, sequence_repo, log_repo, self.event_store)
#
#         self.assertEqual(notification_log_reader[1], item2)
#
#         item3 = 'item3'
#         item4 = 'item4'
#         item5 = 'item5'
#         append_item_to_notification_log(notification_log, item3, sequence_repo, log_repo, self.event_store)
#         append_item_to_notification_log(notification_log, item4, sequence_repo, log_repo, self.event_store)
#         append_item_to_notification_log(notification_log, item5, sequence_repo, log_repo, self.event_store)
#
#         self.assertEqual(notification_log_reader[2], item3)
#         self.assertEqual(notification_log_reader[3], item4)
#         self.assertEqual(notification_log_reader[4], item5)
#
#         self.assertEqual(notification_log_reader[0:2], [item1, item2])
#         self.assertEqual(notification_log_reader[2:4], [item3, item4])
#
#         # Check stop value is handled properly.
#         self.assertEqual(notification_log_reader[2:3], [item3])
#
#         # Index too large.
#         with self.assertRaises(IndexError):
#             notification_log_reader[5]
#
#         # Negative index.
#         with self.assertRaises(IndexError):
#             notification_log_reader[-1]
#
#         # Incomplete slice.
#         with self.assertRaises(IndexError):
#             notification_log_reader[1:]
#
#         # Incomplete slice.
#         with self.assertRaises(IndexError):
#             notification_log_reader[:1]
#
#         # Negative start.
#         with self.assertRaises(IndexError):
#             notification_log_reader[-1:0]
#
#         with self.assertRaises(IndexError):
#             # Slice has negative stop.
#             notification_log_reader[0:-1]
#
#         with self.assertRaises(IndexError):
#             # Slice crosses sequences.
#             notification_log_reader[0:4]
#
#         notification_log_reader[0:1:None]
#         with self.assertRaises(IndexError):
#             # Slice has a step value.
#             notification_log_reader[0:1:1]
#
#         # Add some more items.
#         for i in range(10):
#             item = 'boo{}'.format(i)
#             append_item_to_notification_log(notification_log, item, sequence_repo, log_repo, self.event_store)
#
#             # Check the length.
#             notification_log_index = 5 + i
#             self.assertEqual(len(notification_log_reader), notification_log_index + 1)
#
#             # Check the message.
#             self.assertEqual(notification_log_reader[notification_log_index], item)
#
#
# class TestNotificationLogWithCassandra(WithCassandraActiveRecordStrategies, NotificationLogTestCase):
#     pass
#
#
# class TestNotificationLogWithSQLAlchemy(WithSQLAlchemyActiveRecordStrategies, NotificationLogTestCase):
#     pass



from eventsourcing.tests.example_application_tests.base import ExampleApplicationTestCase
from eventsourcing.tests.sequenced_item_tests.test_cassandra_active_record_strategy import \
    WithCassandraActiveRecordStrategies


class TestExampleApplicationWithCassandra(WithCassandraActiveRecordStrategies, ExampleApplicationTestCase):
    pass



from collections import namedtuple

SequencedItem = namedtuple('SequencedItem', ['sequence_id', 'position', 'topic', 'data'])


class SequencedItemFieldNames(object):

    def __init__(self, sequenced_item_class):
        self._field_names = sequenced_item_class._fields

    @property
    def sequence_id(self):
        # Sequence ID is assumed to be the first field of a sequenced item.
        return self._field_names[0]

    @property
    def position(self):
        # Position is assumed to be the second field of a sequenced item.
        return self._field_names[1]

    @property
    def topic(self):
        # Topic is assumed to be the third field of a sequenced item.
        return self._field_names[2]

    @property
    def data(self):
        # Data is assumed to be the fourth field of a sequenced item.
        return self._field_names[3]

    def __iter__(self):
        return iter(self._field_names)



from abc import ABCMeta, abstractmethod

import six

from eventsourcing.domain.model.events import publish, topic_from_domain_class
from eventsourcing.domain.model.snapshot import AbstractSnapshop, Snapshot
from eventsourcing.infrastructure.eventstore import AbstractEventStore, EventStore
from eventsourcing.infrastructure.sequenceditemmapper import deserialize_domain_entity


class AbstractSnapshotStrategy(six.with_metaclass(ABCMeta)):

    @abstractmethod
    def get_snapshot(self, stored_entity_id, lt=None, lte=None):
        """Returns pre-existing snapshot for stored entity ID from given
        event store, optionally until a particular domain event ID.
        """

    @abstractmethod
    def take_snapshot(self, entity, timestamp):
        """Creates snapshot from given entity, with given domain event ID.
        """


class EventSourcedSnapshotStrategy(AbstractSnapshotStrategy):
    """Snapshot strategy that uses an event sourced snapshot.
    """
    def __init__(self, event_store):
        assert isinstance(event_store, EventStore)
        self.event_store = event_store

    def get_snapshot(self, entity_id, lt=None, lte=None):
        return get_snapshot(entity_id, self.event_store, lt=lt, lte=lte)

    def take_snapshot(self, entity, timestamp=None):
        return take_snapshot(entity, timestamp=timestamp)


def get_snapshot(entity_id, event_store, lt=None, lte=None):
    """
    Get the last snapshot for entity.

    :rtype: Snapshot
    """
    assert isinstance(event_store, AbstractEventStore)
    snapshots = event_store.get_domain_events(entity_id, lt=lt, lte=lte, is_ascending=False, limit=1)
    if len(snapshots) == 1:
        return snapshots[0]


def entity_from_snapshot(snapshot):
    """Deserialises a domain entity from a snapshot object.
    """
    assert isinstance(snapshot, AbstractSnapshop)
    return deserialize_domain_entity(snapshot.topic, snapshot.state)


def take_snapshot(entity, timestamp=None):
    # Make the 'stored entity ID' for the entity, it is used as the Snapshot 'entity ID'.

    # Create the snapshot event.
    snapshot = Snapshot(
        entity_id=entity.id,
        timestamp=timestamp,
        topic=topic_from_domain_class(entity.__class__),
        # Todo: This should be a deepcopy.
        state=entity.__dict__.copy(),
    )
    publish(snapshot)

    # Return the event.
    return snapshot



from uuid import uuid4

from eventsourcing.example.domainmodel import Example
from eventsourcing.infrastructure.eventstore import EventStore
from eventsourcing.infrastructure.sequenceditem import SequencedItem
from eventsourcing.infrastructure.sequenceditemmapper import SequencedItemMapper
from eventsourcing.tests.datastore_tests.test_sqlalchemy import SQLAlchemyDatastoreTestCase
from eventsourcing.tests.sequenced_item_tests.test_sqlalchemy_active_record_strategy import \
    construct_integer_sequenced_active_record_strategy


class TestEventStore(SQLAlchemyDatastoreTestCase):

    def setUp(self):
        super(TestEventStore, self).setUp()
        if self.datastore is not None:
            self.datastore.setup_connection()
            self.datastore.setup_tables()

    def tearDown(self):
        if self.datastore is not None:
            self.datastore.drop_tables()
            self.datastore.drop_connection()
        super(TestEventStore, self).tearDown()

    def construct_event_store(self):
        event_store = EventStore(
            active_record_strategy=construct_integer_sequenced_active_record_strategy(
                datastore=self.datastore,
            ),
            sequenced_item_mapper=SequencedItemMapper(
                sequenced_item_class=SequencedItem,
                event_sequence_id_attr='entity_id',
                event_position_attr='entity_version'
            )
        )
        return event_store

    def test_get_domain_events(self):
        event_store = self.construct_event_store()

        # Check there are zero stored events in the repo.
        entity_id1 = uuid4()
        entity_events = event_store.get_domain_events(entity_id=entity_id1)
        entity_events = list(entity_events)
        self.assertEqual(0, len(entity_events))

        # Check there are zero events in the event store, using iterator.
        entity_events = event_store.get_domain_events(entity_id=entity_id1, page_size=1)
        entity_events = list(entity_events)
        self.assertEqual(0, len(entity_events))

        # Store a domain event.
        event1 = Example.Created(entity_id=entity_id1, a=1, b=2)
        event_store.append(event1)

        # Check there is one event in the event store.
        entity_events = event_store.get_domain_events(entity_id=entity_id1)
        entity_events = list(entity_events)
        self.assertEqual(1, len(entity_events))

        # Check there are two events in the event store, using iterator.
        entity_events = event_store.get_domain_events(entity_id=entity_id1, page_size=1)
        entity_events = list(entity_events)
        self.assertEqual(1, len(entity_events))

        # Store another domain event.
        event1 = Example.AttributeChanged(entity_id=entity_id1, a=1, b=2, entity_version=1)
        event_store.append(event1)

        # Check there are two events in the event store.
        entity_events = event_store.get_domain_events(entity_id=entity_id1)
        entity_events = list(entity_events)
        self.assertEqual(2, len(entity_events))

        # Check there are two events in the event store, using iterator.
        entity_events = event_store.get_domain_events(entity_id=entity_id1, page_size=1)
        entity_events = list(entity_events)
        self.assertEqual(2, len(entity_events))

    def test_get_most_recent_event(self):
        event_store = self.construct_event_store()

        # Check there is no most recent event.
        entity_id1 = uuid4()
        entity_event = event_store.get_most_recent_event(entity_id=entity_id1)
        self.assertEqual(entity_event, None)

        # Store a domain event.
        event1 = Example.Created(entity_id=entity_id1, a=1, b=2)
        event_store.append(event1)

        # Check there is an event.
        entity_event = event_store.get_most_recent_event(entity_id=entity_id1)
        self.assertEqual(entity_event, event1)

    def test_all_domain_events(self):
        event_store = self.construct_event_store()

        # Check there are zero domain events in total.
        domain_events = event_store.all_domain_events()
        domain_events = list(domain_events)
        self.assertEqual(len(domain_events), 0)

        # Store a domain event.
        entity_id1 = uuid4()
        event1 = Example.Created(entity_id=entity_id1, a=1, b=2)
        event_store.append(event1)

        # Store another domain event for the same entity.
        event1 = Example.AttributeChanged(entity_id=entity_id1, a=1, b=2, entity_version=1)
        event_store.append(event1)

        # Store a domain event for a different entity.
        entity_id2 = uuid4()
        event1 = Example.Created(entity_id=entity_id2, a=1, b=2)
        event_store.append(event1)

        # Check there are three domain events in total.
        domain_events = event_store.all_domain_events()
        domain_events = list(domain_events)
        self.assertEqual(len(domain_events), 3)




from eventsourcing.exceptions import ConsistencyError, ProgrammingError
from eventsourcing.utils.time import timestamp_from_uuid

try:
    # Python 3.4+
    from functools import singledispatch
except ImportError:
    from singledispatch import singledispatch

from abc import ABCMeta, abstractmethod
from inspect import isfunction
from six import with_metaclass

from eventsourcing.domain.model.events import publish, QualnameABCMeta, AggregateEvent


class EntityIDConsistencyError(ConsistencyError):
    pass


class EntityVersionConsistencyError(ConsistencyError):
    pass


class CreatedMutatorRequiresTypeNotInstance(ConsistencyError):
    pass


class EntityIsDiscarded(AssertionError):
    pass


class Created(AggregateEvent):
    def __init__(self, entity_version=0, **kwargs):
        super(Created, self).__init__(entity_version=entity_version, **kwargs)


class AttributeChanged(AggregateEvent):
    pass


class Discarded(AggregateEvent):
    pass


# Todo: Decompose this, to allow for entities without timestamps and without versions.
class DomainEntity(with_metaclass(QualnameABCMeta)):

    def __init__(self, entity_id):
        self._id = entity_id
        self._is_discarded = False

    def _assert_not_discarded(self):
        if self._is_discarded:
            raise EntityIsDiscarded("Entity is discarded")

    @property
    def id(self):
        return self._id

    def _validate_originator(self, event):
        self._validate_originator_id(event)

    def _validate_originator_id(self, event):
        """
        Checks the event's entity ID matches this entity's ID.
        """
        if self._id != event.entity_id:
            raise EntityIDConsistencyError(
                "Entity ID '{}' not equal to event's entity ID '{}'"
                "".format(self.id, event.entity_id)
            )

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def _change_attribute(self, name, value):
        self._assert_not_discarded()
        event_class = getattr(self, 'AttributeChanged', AttributeChanged)
        event = event_class(name=name, value=value, entity_id=self._id, entity_version=self._version)
        self._apply(event)
        publish(event)

    def discard(self):
        self._assert_not_discarded()
        event_class = getattr(self, 'Discarded', Discarded)
        event = event_class(entity_id=self._id, entity_version=self._version)
        self._apply(event)
        publish(event)

    def _apply(self, event):
        self.mutate(event=event, entity=self)

    @classmethod
    def mutate(cls, entity=None, event=None):
        initial = entity if entity is not None else cls
        return cls._mutator(event, initial)

    @staticmethod
    def _mutator(event, initial):
        return entity_mutator(event, initial)


class VersionedEntity(DomainEntity):

    def __init__(self, entity_version=None, **kwargs):
        super(VersionedEntity, self).__init__(**kwargs)
        self._version = entity_version

    @property
    def version(self):
        return self._version

    def _increment_version(self):
        if self._version is not None:
            self._version += 1

    def _validate_originator(self, event):
        super(VersionedEntity, self)._validate_originator(event)
        self._validate_originator_version(event)

    def _validate_originator_version(self, event):
        """
        Checks the event's entity version matches this entity's version.
        """
        if self._version != event.entity_version:
            raise EntityVersionConsistencyError(
                ("Event version '{}' not equal to entity version '{}', "
                 "event type: '{}', entity type: '{}', entity ID: '{}'"
                 "".format(event.entity_version, self._version,
                           type(event).__name__, type(self).__name__, self._id)
                 )
            )


class TimestampedEntity(DomainEntity):
    def __init__(self, timestamp=None, **kwargs):
        super(TimestampedEntity, self).__init__(**kwargs)
        self._created_on = timestamp
        self._last_modified_on = timestamp

    @property
    def created_on(self):
        return self._created_on

    @property
    def last_modified_on(self):
        return self._last_modified_on


class TimeuuidedEntity(DomainEntity):
    def __init__(self, event_id=None, **kwargs):
        super(TimeuuidedEntity, self).__init__(**kwargs)
        self._initial_event_id = event_id
        self._last_event_id = event_id

    @property
    def created_on(self):
        return timestamp_from_uuid(self._initial_event_id)

    @property
    def last_modified_on(self):
        return timestamp_from_uuid(self._last_event_id)


class TimestampedVersionedEntity(TimestampedEntity, VersionedEntity):
    pass


class TimeuuidedVersionedEntity(TimeuuidedEntity, VersionedEntity):
    pass


@singledispatch
def entity_mutator(event, _):
    raise NotImplementedError("Event type not supported: {}".format(type(event)))


@entity_mutator.register(Created)
def created_mutator(event, cls):
    assert isinstance(event, Created), event
    if not isinstance(cls, type):
        msg = ("Mutator for Created event requires entity type not instance: {} "
               "(event entity id: {}, event type: {})"
               "".format(type(cls), event.entity_id, type(event)))
        raise CreatedMutatorRequiresTypeNotInstance(msg)
    assert issubclass(cls, TimestampedVersionedEntity), cls
    try:
        self = cls(**event.__dict__)
    except TypeError as e:
        raise TypeError("Class {} {}. Given {} from event type {}".format(cls, e, event.__dict__, type(event)))
    self._increment_version()
    return self


@entity_mutator.register(AttributeChanged)
def attribute_changed_mutator(event, self):
    assert isinstance(self, TimestampedVersionedEntity), self
    self._validate_originator(event)
    setattr(self, event.name, event.value)
    self._last_modified_on = event.timestamp
    self._increment_version()
    return self


@entity_mutator.register(Discarded)
def discarded_mutator(event, self):
    assert isinstance(self, TimestampedVersionedEntity), self
    self._validate_originator(event)
    self._is_discarded = True
    self._increment_version()
    return None


def attribute(getter):
    """
    When used as a method decorator, returns a property object
    with the method as the getter and a setter defined to call
    instance method _change_attribute(), which publishes an
    AttributeChanged event.
    """
    if isfunction(getter):
        def setter(self, value):
            assert isinstance(self, TimestampedVersionedEntity), type(self)
            name = '_' + getter.__name__
            self._change_attribute(name=name, value=value)

        def new_getter(self):
            assert isinstance(self, TimestampedVersionedEntity), type(self)
            name = '_' + getter.__name__
            return getattr(self, name)

        return property(fget=new_getter, fset=setter)
    else:
        raise ProgrammingError("Expected a function, got: {}".format(repr(getter)))


class AbstractEntityRepository(with_metaclass(ABCMeta)):
    @abstractmethod
    def __getitem__(self, entity_id):
        """
        Returns entity for given ID.
        """

    @abstractmethod
    def __contains__(self, entity_id):
        """
        Returns True or False, according to whether or not entity exists.
        """


class Aggregate(TimestampedVersionedEntity):
    """
    For aggregates in Domain Driven Design.
    """


class AggregateRepository(AbstractEntityRepository):
    """
    For aggregate repositories in Domain Driven Design.
    """



import unittest

from eventsourcing.application.policies import CombinedPersistencePolicy, TimestampedEntityEvent
from eventsourcing.domain.model.events import VersionedEntityEvent, publish
from eventsourcing.infrastructure.eventstore import AbstractEventStore

try:
    from unittest import mock
except:
    import mock


class TestCombinedPersistencePolicy(unittest.TestCase):
    def setUp(self):
        self.ve_es = mock.Mock(spec=AbstractEventStore)
        self.te_es = mock.Mock(spec=AbstractEventStore)
        self.policy = CombinedPersistencePolicy(
            versioned_entity_event_store=self.ve_es,
            timestamped_entity_event_store=self.te_es,
        )

    def tearDown(self):
        self.policy.close()

    def test_published_events_are_appended_to_event_store(self):
        # Check the event store's append method has NOT been called.
        assert isinstance(self.ve_es, AbstractEventStore)
        assert isinstance(self.te_es, AbstractEventStore)
        self.assertEqual(0, self.ve_es.append.call_count)
        self.assertEqual(0, self.te_es.append.call_count)

        # Publish a (mock) versioned entity event.
        domain_event1 = mock.Mock(spec=VersionedEntityEvent)
        publish(domain_event1)

        # Check the append method HAS been called once with the domain event.
        self.ve_es.append.assert_called_once_with(domain_event1)
        self.assertEqual(0, self.te_es.append.call_count)

        # Publish a (mock) timestamp entity event.
        domain_event2 = mock.Mock(spec=TimestampedEntityEvent)
        publish(domain_event2)

        # Check the append method HAS been called once with the domain event.
        self.ve_es.append.assert_called_once_with(domain_event1)
        self.te_es.append.assert_called_once_with(domain_event2)



# import datetime
# import json
# import os
# import traceback
# from multiprocessing.pool import Pool
# from time import sleep
# from uuid import uuid1, uuid4
#
# import six
#
# from eventsourcing.exceptions import ConcurrencyError, DatasourceOperationError
# from eventsourcing.infrastructure.datastore.cassandraengine import CassandraDatastore, CassandraSettings
# from eventsourcing.infrastructure.datastore.sqlalchemyorm import SQLAlchemyDatastore, SQLAlchemySettings
# from eventsourcing.infrastructure.eventstore import AbstractStoredEventRepository
# from eventsourcing.infrastructure.storedevents.cassandrarepo import CassandraStoredEventRepository, CqlStoredEvent
# from eventsourcing.infrastructure.storedevents.pythonobjectsrepo import PythonObjectsStoredEventRepository
# from eventsourcing.infrastructure.storedevents.sqlalchemyrepo import SQLAlchemyStoredEventRepository
# from eventsourcing.infrastructure.transcoding import StoredEvent
# from eventsourcing.tests.base import notquick
# from eventsourcing.tests.datastore_tests.test_cassandra import DEFAULT_KEYSPACE_FOR_TESTING
# from eventsourcing.tests.sequenced_item_tests.base import WithActiveRecordStrategies
# from eventsourcing.tests.sequenced_item_tests.test_cassandra_active_record_strategy import \
#     WithCassandraActiveRecordStrategies
# from eventsourcing.tests.sequenced_item_tests.test_sqlalchemy_active_record_strategy import \
#     WithSQLAlchemyActiveRecordStrategies
#
#
# class OptimisticConcurrencyControlTestCase(WithActiveRecordStrategies):
#     @notquick()
#     def test_optimistic_concurrency_control(self):
#         """Appends lots of events, but with a pool of workers
#         all trying to add the same sequence of events.
#         """
#         # Start a pool.
#         pool_size = 3
#         print("Pool size: {}".format(pool_size))
#
#         # Erm, this is only needed for SQLite database file.
#         # Todo: Maybe factor out a 'get_initargs()' method on this class,
#         # so this detail is localised to the test cases that need it.
#         if hasattr(self, 'temp_file'):
#             temp_file_name = getattr(self, 'temp_file').name
#         else:
#             temp_file_name = None
#
#         pool = Pool(
#             initializer=pool_initializer,
#             processes=pool_size,
#             initargs=(type(self.integer_sequence_active_record_strategy), temp_file_name),
#         )
#
#         # Append duplicate events to the repo, or at least try...
#         number_of_events = 40
#         self.assertGreater(number_of_events, pool_size)
#         stored_entity_id = uuid4().hex
#         sequence_of_args = [(number_of_events, stored_entity_id)] * pool_size
#         results = pool.map(append_lots_of_events_to_repo, sequence_of_args)
#         total_successes = []
#         total_failures = []
#         for result in results:
#             if isinstance(result, Exception):
#                 print(result.args[0][1])
#                 raise result
#             else:
#                 successes, failures = result
#                 assert isinstance(successes, list), result
#                 assert isinstance(failures, list), result
#                 total_successes.extend(successes)
#                 total_failures.extend(failures)
#
#         # Close the pool.
#         pool.close()
#
#         # Check each event version was written exactly once.
#         self.assertEqual(sorted([i[0] for i in total_successes]), list(range(number_of_events)))
#
#         # Check there was contention that caused at least one concurrency error.
#         set_failures = set([i[0] for i in total_failures])
#         self.assertTrue(len(set_failures))
#
#         # Check each child wrote at least one event.
#         self.assertEqual(len(set([i[1] for i in total_successes])), pool_size)
#
#         # Check each child encountered at least one concurrency error.
#         self.assertEqual(len(set([i[1] for i in total_failures])), pool_size)
#
#         # Check the repo actually has a contiguous version sequence.
#         events = self.integer_sequence_active_record_strategy.get_stored_events(stored_entity_id)
#         self.assertEqual(len(events), number_of_events)
#         version_counter = 0
#         for event in events:
#             assert isinstance(event, StoredEvent)
#             attr_values = json.loads(event.event_attrs)
#             self.assertEqual(attr_values['entity_version'], version_counter)
#             version_counter += 1
#
#         # Join the pool.
#         pool.join()
#
#     @staticmethod
#     def append_lots_of_events_to_repo(args):
#
#         num_events_to_create, stored_entity_id = args
#
#         success_count = 0
#         assert isinstance(num_events_to_create, six.integer_types)
#
#         successes = []
#         failures = []
#
#         try:
#
#             while True:
#                 # Imitate an entity getting refreshed, by getting the version of the last event.
#                 assert isinstance(worker_repo, AbstractStoredEventRepository)
#                 events = worker_repo.get_stored_events(stored_entity_id, limit=1, query_ascending=False)
#                 if len(events):
#                     current_version = json.loads(events[0].event_attrs)['entity_version']
#                     new_version = current_version + 1
#                 else:
#                     current_version = None
#                     new_version = 0
#
#                 # Stop before the version number gets too high.
#                 if new_version >= num_events_to_create:
#                     break
#
#                 pid = os.getpid()
#                 try:
#
#                     # Append an event.
#                     stored_event = StoredEvent(
#                         event_id=uuid1().hex,
#                         stored_entity_id=stored_entity_id,
#                         event_topic='topic',
#                         event_attrs=json.dumps({'a': 1, 'b': 2, 'entity_version': new_version}),
#                     )
#                     started = datetime.datetime.now()
#                     worker_repo.append(
#                         new_stored_event=stored_event,
#                         new_version_number=new_version,
#                         max_retries=10,
#                         artificial_failure_rate=0.25,
#                     )
#                 except ConcurrencyError:
#                     # print("PID {} got concurrent exception writing event at version {} at {}".format(
#                     #     pid, new_version, started, datetime.datetime.now() - started))
#                     failures.append((new_version, pid))
#                     sleep(0.01)
#                 except DatasourceOperationError:
#                     # print("PID {} got concurrent exception writing event at version {} at {}".format(
#                     #     pid, new_version, started, datetime.datetime.now() - started))
#                     # failures.append((new_version, pid))
#                     sleep(0.01)
#                 else:
#                     print("PID {} success writing event at version {} at {} in {}".format(
#                         pid, new_version, started, datetime.datetime.now() - started))
#                     success_count += 1
#                     successes.append((new_version, pid))
#                     # Delay a successful writer, to give other processes a chance to write the next event.
#                     sleep(0.03)
#
#         # Return to parent process the successes and failure, or an exception.
#         except Exception as e:
#             msg = traceback.format_exc()
#             print(" - failed to append event: {}".format(msg))
#             return Exception((e, msg))
#         else:
#             return (successes, failures)
#
#
# @notquick()
# class TestOptimisticConcurrencyControlWithCassandra(WithCassandraActiveRecordStrategies,
#                                                     OptimisticConcurrencyControlTestCase):
#     pass
#
#
# class TestOptimisticConcurrencyControlWithSQLAlchemy(WithSQLAlchemyActiveRecordStrategies,
#                                                      OptimisticConcurrencyControlTestCase):
#     use_named_temporary_file = True
#
#
# worker_repo = None
#
#
# def pool_initializer(stored_repo_class, temp_file_name):
#     global worker_repo
#     worker_repo = construct_repo_for_worker(stored_repo_class, temp_file_name)
#
#
# def construct_repo_for_worker(stored_repo_class, temp_file_name):
#     if stored_repo_class is CassandraStoredEventRepository:
#         datastore = CassandraDatastore(
#             settings=CassandraSettings(default_keyspace=DEFAULT_KEYSPACE_FOR_TESTING),
#             tables=(CqlStoredEvent,)
#         )
#         datastore.drop_connection()
#         datastore.setup_connection()
#         repo = CassandraStoredEventRepository(
#             stored_event_table=CqlStoredEvent,
#             always_check_expected_version=True,
#             always_write_entity_version=True,
#         )
#     elif stored_repo_class is SQLAlchemyStoredEventRepository:
#         uri = 'sqlite:///' + temp_file_name
#         datastore = SQLAlchemyDatastore(
#             settings=SQLAlchemySettings(uri=uri),
#             tables=(SqlStoredEvent,),
#         )
#         datastore.setup_connection()
#         repo = SQLAlchemyStoredEventRepository(
#             datastore=datastore,
#             stored_event_table=SqlStoredEvent,
#             always_check_expected_version=True,
#             always_write_entity_version=True,
#         )
#     elif stored_repo_class is PythonObjectsStoredEventRepository:
#         repo = PythonObjectsStoredEventRepository(
#             always_check_expected_version=True,
#             always_write_entity_version=True,
#         )
#     else:
#         raise Exception("Stored repo class not yet supported in test: {}".format(stored_repo_class))
#     return repo
#
#
# def append_lots_of_events_to_repo(args):
#     return OptimisticConcurrencyControlTestCase.append_lots_of_events_to_repo(args)



# import six
#
# from eventsourcing.domain.model.timebucketedlog import TimebucketedlogRepository
# from eventsourcing.domain.model.notificationlog import NotificationLog
# from eventsourcing.domain.model.sequence import Sequence, SequenceRepository
# from eventsourcing.exceptions import SequenceFullError
# from eventsourcing.infrastructure.eventstore import AbstractEventStore
# from eventsourcing.infrastructure.timebucketedlog_reader import TimebucketedlogReader
# from eventsourcing.infrastructure.sequence import SequenceReader, append_item_to_sequence
#
#
# def append_item_to_notification_log(notification_log, item, sequence_repo, log_repo, event_store):
#     """Appends item to current sequence, unless the sequence is full.
#
#     If the sequence is full, the sequence number is incremented and
#     the item is added to the next sequence."""
#     assert isinstance(notification_log, NotificationLog), notification_log
#     assert isinstance(sequence_repo, SequenceRepository)
#     assert isinstance(log_repo, TimebucketedlogRepository)
#     # Get the sequence.
#     current_sequence = get_current_notification_log_sequence(notification_log, sequence_repo, log_repo, event_store)
#     assert isinstance(current_sequence, Sequence)
#     try:
#         append_item_to_sequence(current_sequence.name, item, sequence_repo.event_player,
#                                 notification_log.sequence_size)
#     except SequenceFullError:
#         # Roll over the sequence.
#         # - construct next sequence ID
#         current_sequence_number = int(current_sequence.name.split('::')[-1])
#         next_sequence_id = make_notification_log_sequence_id(notification_log.name, current_sequence_number + 1)
#         # - write message into time-bucketed log
#         log = log_repo.get_or_create(notification_log.name, notification_log.bucket_size)
#         log.append_message(next_sequence_id)
#         next_sequence = sequence_repo.get_or_create(next_sequence_id)
#
#         append_item_to_sequence(next_sequence.name, item, sequence_repo.event_player, notification_log.sequence_size)
#
#
# def get_current_notification_log_sequence(notification_log, sequence_repo, log_repo, event_store):
#     # Get time-bucketed log.
#     sequence_id = get_current_notification_log_sequence_id(notification_log, sequence_repo, log_repo, event_store)
#     return sequence_repo.get_or_create(sequence_id)
#
#
# def get_current_notification_log_sequence_id(notification_log, sequence_repo, log_repo, event_store):
#     assert isinstance(notification_log, NotificationLog), notification_log
#     assert isinstance(sequence_repo, SequenceRepository), sequence_repo
#     assert isinstance(log_repo, TimebucketedlogRepository), log_repo
#     log = log_repo.get_or_create(notification_log.name, notification_log.bucket_size)
#     log_reader = TimebucketedlogReader(log, event_store)
#     events = list(log_reader.get_events(limit=1))
#     if len(events):
#         most_recent_event = events[0]
#         sequence_id = most_recent_event.message
#     else:
#         sequence_id = make_notification_log_sequence_id(notification_log.name, 0)
#     return sequence_id
#
#
# def make_notification_log_sequence_id(notification_log_name, sequence_number):
#     assert isinstance(notification_log_name, six.string_types)
#     assert isinstance(sequence_number, six.integer_types)
#     return '{}::{}'.format(notification_log_name, sequence_number)
#
#
# class NotificationLogReader(object):
#     def __init__(self, notification_log, sequence_repo, log_repo, event_store):
#         assert isinstance(notification_log, NotificationLog), notification_log
#         assert isinstance(sequence_repo, SequenceRepository)
#         assert isinstance(log_repo, TimebucketedlogRepository)
#         assert isinstance(event_store, AbstractEventStore), event_store
#         self.notification_log = notification_log
#         self.sequence_repo = sequence_repo
#         self.log_repo = log_repo
#         self.event_store = event_store
#
#     def __getitem__(self, item):
#         assert isinstance(item, (six.integer_types, slice))
#         sequence_size = self.notification_log.sequence_size
#         sequence_number = None
#         sequence_item = None
#
#         if isinstance(item, six.integer_types):
#             # Get a single item from a sequence.
#             if item < 0:
#                 raise IndexError('Negative index values not yet supported by notification log')
#             sequence_number, sequence_item = divmod(item, sequence_size)
#         elif isinstance(item, slice):
#             # Get a number of items from part of a sequence.
#             if item.start is None or item.stop is None:
#                 raise IndexError('Slices without both start and stop values are not supported')
#             if item.start < 0 or item.stop < 0:
#                 raise IndexError('Negative index values not yet supported by notification log')
#             if item.step is not None:
#                 raise IndexError('Slices with a step value are not supported')
#
#             # Identify sequence numbers and sequence index values from start and stop.
#             start_sequence_number, slice_start = divmod(item.start, sequence_size)
#             stop_sequence_number = (item.stop - 1) // sequence_size
#
#             # Check everything will come from a single sequence.
#             if start_sequence_number != stop_sequence_number:
#                 raise IndexError("Slice crosses sequence objects ({}, {}), which isn't supported yet".format(
#                     start_sequence_number, stop_sequence_number
#                 ))
#
#             # Remember the sequence number.
#             sequence_number = start_sequence_number
#
#             # Calculate sequence slice stop value.
#             index_offset = sequence_number * sequence_size
#             slice_stop = item.stop - index_offset
#             sequence_item = slice(slice_start, slice_stop)
#
#         # Make a sequence ID from the notification log name and the sequence number.
#         sequence_id = make_notification_log_sequence_id(self.notification_log.name, sequence_number)
#
#         # Read item(s) from the sequence.
#         sequence = self.sequence_repo.get_or_create(sequence_id)
#         reader = SequenceReader(sequence, self.sequence_repo.event_player)
#         try:
#             return reader[sequence_item]
#         except IndexError:
#             msg = "Notification log item {} not found (item {} in sequence {})".format(item, sequence_item,
#                                                                                        sequence_number)
#             raise IndexError(msg)
#
#     def __len__(self):
#         # Get the current sequence.
#         current_sequence = get_current_notification_log_sequence(self.notification_log,
#                                                                  self.sequence_repo,
#                                                                  self.log_repo,
#                                                                  self.event_store)
#         assert isinstance(current_sequence, Sequence)
#         sequence_number = int(current_sequence.name.split('::')[-1])
#         event = self.sequence_repo.event_player.get_most_recent_event(current_sequence.name)
#         sequence_count = event.entity_version
#         return (sequence_number * self.notification_log.sequence_size) + sequence_count



from collections import namedtuple
from uuid import UUID

from sqlalchemy.sql.schema import Column
from sqlalchemy.sql.sqltypes import Float, String

from eventsourcing.application.policies import PersistencePolicy
from eventsourcing.example.domainmodel import register_new_example
from eventsourcing.example.infrastructure import ExampleRepository
from eventsourcing.infrastructure.eventstore import EventStore
from eventsourcing.infrastructure.sequenceditemmapper import SequencedItemMapper
from eventsourcing.infrastructure.sqlalchemy.activerecords import SQLAlchemyActiveRecordStrategy, \
    SqlIntegerSequencedItem
from eventsourcing.infrastructure.sqlalchemy.datastore import Base, SQLAlchemyDatastore, SQLAlchemySettings
from eventsourcing.tests.datastore_tests.base import AbstractDatastoreTestCase

# This tests extending the sequenced item class with some more fields. How easy is it?
# Just needed to define the extended type, define a suitable active record
# class, and extend the sequenced itemevent mapper to derive values for the
# extra attributes. It's easy.

# Define the sequenced item class.
ExtendedSequencedItem = namedtuple('ExtendedSequencedItem',
                                   ['sequence_id', 'position', 'topic', 'data', 'timestamp', 'event_type'])


# Extend the database table definition to support the extra fields.
class SqlExtendedIntegerSequencedItem(SqlIntegerSequencedItem):
    # Timestamp of the event.
    timestamp = Column(Float())

    # Type of the event (class name).
    event_type = Column(String(100))


# Extend the sequenced item mapper to derive the extra values.
class ExtendedSequencedItemMapper(SequencedItemMapper):
    def construct_item_args(self, domain_event):
        args = super(ExtendedSequencedItemMapper, self).construct_item_args(domain_event)
        timestamp = domain_event.timestamp
        event_type = domain_event.__class__.__qualname__
        return args + (timestamp, event_type)


# Define an application object.
class ExampleApplicationWithExtendedSequencedItemType(object):
    def __init__(self, datastore):
        self.event_store = EventStore(
            active_record_strategy=SQLAlchemyActiveRecordStrategy(
                datastore=datastore,
                active_record_class=SqlExtendedIntegerSequencedItem,
                sequenced_item_class=ExtendedSequencedItem,
            ),
            sequenced_item_mapper=ExtendedSequencedItemMapper(
                sequenced_item_class=ExtendedSequencedItem,
                event_sequence_id_attr='entity_id',
                event_position_attr='entity_version',
            )
        )
        self.repository = ExampleRepository(
            event_store=self.event_store,
        )
        self.persistence_policy = PersistencePolicy(self.event_store)

    def close(self):
        self.persistence_policy.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class TestExampleWithExtendedSequencedItemType(AbstractDatastoreTestCase):
    def setUp(self):
        super(TestExampleWithExtendedSequencedItemType, self).setUp()
        self.datastore.setup_connection()
        self.datastore.setup_tables()

    def tearDown(self):
        self.datastore.drop_tables()
        self.datastore.drop_connection()
        super(TestExampleWithExtendedSequencedItemType, self).setUp()

    def construct_datastore(self):
        return SQLAlchemyDatastore(
            base=Base,
            settings=SQLAlchemySettings(),
            tables=(SqlExtendedIntegerSequencedItem,),
        )

    def test(self):
        with ExampleApplicationWithExtendedSequencedItemType(self.datastore) as app:
            # Create entity.
            entity1 = register_new_example(a='a', b='b')
            self.assertIsInstance(entity1.id, UUID)
            self.assertEqual(entity1.a, 'a')
            self.assertEqual(entity1.b, 'b')

            # Check there is a stored event.
            all_records = list(app.event_store.active_record_strategy.filter())
            self.assertEqual(len(all_records), 1)
            active_record = all_records[0]
            self.assertEqual(active_record.sequence_id, entity1.id)
            self.assertEqual(active_record.position, 0)
            self.assertEqual(active_record.event_type, 'Example.Created', active_record.event_type)
            self.assertEqual(active_record.timestamp, entity1.created_on)

            # Read entity from repo.
            retrieved_obj = app.repository[entity1.id]
            self.assertEqual(retrieved_obj.id, entity1.id)



from eventsourcing.domain.model.collection import AbstractCollectionRepository, Collection
from eventsourcing.infrastructure.eventsourcedrepository import EventSourcedRepository


class CollectionRepository(EventSourcedRepository, AbstractCollectionRepository):
    """
    Event sourced repository for the Collection domain model entity.
    """
    mutator = Collection.mutate



from uuid import uuid4

from eventsourcing.domain.model.sequence import Sequence, start_sequence
from eventsourcing.exceptions import SequenceFullError
from eventsourcing.infrastructure.event_sourced_repos.sequence import SequenceRepo
from eventsourcing.infrastructure.sequence import SequenceReader, append_item_to_sequence
from eventsourcing.tests.sequenced_item_tests.test_cassandra_active_record_strategy import \
    WithCassandraActiveRecordStrategies
from eventsourcing.tests.sequenced_item_tests.test_sqlalchemy_active_record_strategy import \
    WithSQLAlchemyActiveRecordStrategies
from eventsourcing.tests.sequenced_item_tests.base import WithPersistencePolicy


class SequenceTestCase(WithPersistencePolicy):
    def test(self):
        repo = SequenceRepo(self.versioned_entity_event_store)

        # Start a new sequence.
        name = uuid4()

        # Check get_or_create() can create a new sequence.
        sequence = repo.get_or_create(name)
        self.assertIsInstance(sequence, Sequence)
        self.assertEqual(sequence.name, name)

        # Check get_or_create() can return an existing sequence.
        sequence = repo.get_or_create(name)
        self.assertIsInstance(sequence, Sequence)
        self.assertEqual(sequence.name, name)

        # Append some items.
        append_item_to_sequence(name, 'item1', repo.event_player)
        append_item_to_sequence(name, 'item2', repo.event_player)
        append_item_to_sequence(name, 'item3', repo.event_player)

        # Check the sequence in the repo.
        sequence = repo[name]
        self.assertIsInstance(sequence, Sequence)
        self.assertEqual(sequence.name, name)
        self.assertEqual(sequence.version, 1)

        # Check the sequence indexing.
        reader = SequenceReader(sequence, repo.event_player)
        self.assertEqual(reader[0], 'item1')
        self.assertEqual(reader[-3], 'item1')
        self.assertEqual(reader[1], 'item2')
        self.assertEqual(reader[-2], 'item2')
        self.assertEqual(reader[2], 'item3')
        self.assertEqual(reader[-1], 'item3')

        # Check slices also work.
        self.assertEqual(reader[0:3], ['item1', 'item2', 'item3'])
        self.assertEqual(reader[0:2], ['item1', 'item2'])
        self.assertEqual(reader[0:-1], ['item1', 'item2'])
        self.assertEqual(reader[0:1], ['item1'])
        self.assertEqual(reader[0:-2], ['item1'])
        self.assertEqual(reader[1:3], ['item2', 'item3'])
        self.assertEqual(reader[1:2], ['item2'])
        self.assertEqual(reader[-2:-1], ['item2'])
        self.assertEqual(reader[1:1], [])
        self.assertEqual(reader[-2:-2], [])
        self.assertEqual(reader[2:3], ['item3'])
        self.assertEqual(reader[3:3], [])
        self.assertEqual(reader[0:300], ['item1', 'item2', 'item3'])
        self.assertEqual(reader[2:1], [])
        self.assertEqual(reader[2:-2], [])

        self.assertEqual(reader[0:], ['item1', 'item2', 'item3'])
        self.assertEqual(reader[1:], ['item2', 'item3'])
        self.assertEqual(reader[2:], ['item3'])
        self.assertEqual(reader[3:], [])
        self.assertEqual(reader[4:], [])
        self.assertEqual(reader[-1:], ['item3'])
        self.assertEqual(reader[-2:], ['item2', 'item3'])
        self.assertEqual(reader[-3:], ['item1', 'item2', 'item3'])
        self.assertEqual(reader[-4:], ['item1', 'item2', 'item3'])

        self.assertEqual(reader[:0], [])
        self.assertEqual(reader[:1], ['item1'])
        self.assertEqual(reader[:2], ['item1', 'item2'])
        self.assertEqual(reader[:3], ['item1', 'item2', 'item3'])
        self.assertEqual(reader[:4], ['item1', 'item2', 'item3'])
        self.assertEqual(reader[:-1], ['item1', 'item2'])
        self.assertEqual(reader[:-2], ['item1'])
        self.assertEqual(reader[:-3], [])
        self.assertEqual(reader[:-4], [])

        # Check iterator.
        for i, item in enumerate(reader):
            self.assertEqual(item, 'item{}'.format(i + 1))

        # Check len.
        self.assertEqual(len(reader), 3)

        # Check index errors.
        # - out of range
        with self.assertRaises(IndexError):
            # noinspection PyStatementEffect
            reader[3]

        with self.assertRaises(IndexError):
            # noinspection PyStatementEffect
            reader[-4]

        with self.assertRaises(SequenceFullError):
            # Append another item.
            append_item_to_sequence(name, 'item1', repo.event_player, max_size=1)


class TestCassandraSequence(WithCassandraActiveRecordStrategies, SequenceTestCase):
    pass


class TestSQLAlchemySequence(WithSQLAlchemyActiveRecordStrategies, SequenceTestCase):
    pass



# coding=utf-8
from __future__ import unicode_literals

from collections import OrderedDict
from time import sleep
from uuid import uuid4

import six
from singledispatch import singledispatch

from eventsourcing.domain.model.collection import Collection
from eventsourcing.domain.model.entity import AttributeChanged, Created, Discarded, AbstractEntityRepository, \
    Aggregate, attribute, entity_mutator
from eventsourcing.domain.model.events import publish, AggregateEvent
from eventsourcing.exceptions import ConcurrencyError, RepositoryKeyError

# Use a private code point to terminate the strings.
STRING_ID_END = '\uEFFF'


class GeneralizedSuffixTree(Aggregate):
    """A suffix tree for string matching. Uses Ukkonen's algorithm
    for construction.

    Adding strings to individual suffix trees needs to be serialized,
    so you need to use locks around calls to add_string() in a
    distributed application.
    """

    class Created(Created):
        pass

    class AttributeChanged(AttributeChanged):
        pass

    class Discarded(Discarded):
        pass

    def __init__(self, root_node_id, case_insensitive=False, **kwargs):
        super(GeneralizedSuffixTree, self).__init__(**kwargs)
        self._root_node_id = root_node_id
        self._case_insensitive = case_insensitive
        # self._nodes = {}
        self._string = None
        self._suffix = None
        self._node_repo = None
        self._node_child_collection_repo = None
        self._edge_repo = None
        self._stringid_collection_repo = None

    @attribute
    def string(self):
        return self._string

    @property
    def nodes(self):
        return self._nodes

    @property
    def root_node_id(self):
        return self._root_node_id

    @property
    def suffix(self):
        return self._suffix

    @property
    def case_insensitive(self):
        return self._case_insensitive

    def __repr__(self):
        """
        Lists edges in the suffix tree
        """
        # Todo: Something that shows the actual tree.
        return ('GeneralizedSuffixTree(id={0.id}, root_node_id={0.root_node_id}, case_insensitive={'
               '0.case_insensitive})'.format(self))

    def add_string(self, string, string_id):
        assert isinstance(string, six.string_types)
        assert isinstance(string_id, six.string_types)

        # Lower the string, if it's a case insensitive tree.
        if self._case_insensitive:
            string = string.lower()
        assert STRING_ID_END not in string_id
        assert STRING_ID_END not in string

        # Append the (hopefully unique) string ID to the string.
        string_extended = u"{}{}".format(string, STRING_ID_END)

        self._suffix = Suffix(self._root_node_id, 0, -1)
        for i in range(len(string_extended)):
            self._add_prefix(i, string_extended, string_id)

    def _add_prefix(self, i, string, string_id):
        """The core construction method.
        """
        last_parent_node_id = None
        parent_node_id = None
        added_node_ids = set()
        while True:

            parent_node_id = self.suffix.source_node_id

            # Break if prefix is already in tree.
            if self.suffix.explicit():
                # Explicit suffix points to an existing node. Try to find the node.
                prefix = string[i]
                try:
                    edge = self.get_edge(make_edge_id(parent_node_id, prefix))
                except KeyError:
                    # If the edge doesn't exist, the prefix isn't in the
                    # tree, so continue by adding an edge and a node.
                    # print("Explicit suffix, edge not found for letter '{}'".format(prefix))
                    pass
                else:
                    # The edge was found, so exit the loop.
                    # print("Explicit suffix, edge found for letter '{}': {}".format(prefix, edge))

                    # Although, might need to add string ID to node collection, if it's a leaf and exact match.
                    dest_node = self.get_node(edge.dest_node_id)
                    if self.is_leaf(dest_node):

                        # If it's an exact match...
                        if edge.label == string[i:]:

                            # Add the string ID to the node collection.
                            self.add_string_id(edge.dest_node_id, string_id)
                            # print("Added string ID '{}' to explicit exact match {}".format(string_id, dest_node))

                            # Attempt to replicate for a duplicate what happens first time a string was added.
                            if edge.dest_node_id not in added_node_ids:
                                # Adjust the active suffix.
                                self.follow_suffix_link_or_increment_suffix_first_char(string)

                    break

            else:
                # Break if prefix is already in tree.
                prefix = string[self.suffix.first_char_index]
                edge = self.get_edge(make_edge_id(self.suffix.source_node_id, prefix))
                suffix_label_offset = self.suffix.length + 1
                if suffix_label_offset > len(edge.label) - 1:
                    raise ValueError(
                        "Suffix offset too large for label: %s: %s" % (suffix_label_offset, edge.label[:100]))
                # print("Implicit edge dest node: {}".format(edge.dest_node_id))
                if edge.label[suffix_label_offset] == string[i]:
                    # print("Matched edge label til end of suffix: {} {}".format(edge.label[:100],
                    # suffix_label_offset))
                    # The suffix was found in this edge, so exit the loop.

                    # Add the string ID to the node collection.
                    self.add_string_id(edge.dest_node_id, string_id)
                    # print("Added string ID '{}' to implicit exact match: {}".format(string_id, self.get_node(
                    # edge.dest_node_id)))

                    break

                else:
                    # Split the edge, with a new middle node that will be the parent node.
                    parent_node_id = self._split_edge(edge, self.suffix)

            # Otherwise the prefix isn't in the tree, so create a node and an edge for this prefix.
            new_leaf_node = register_new_node(string_id=string_id, is_leaf=True)

            # Remeber this node was created in this call.
            added_node_ids.add(new_leaf_node.id)

            # Make a new edge from the parent node to the
            # new leaf node. Make a label for the edge to
            # the leaf, using the suffix of the string
            # from i onwards.
            new_edge = register_new_edge(
                edge_id=make_edge_id(parent_node_id, string[i]),
                label=string[i:],
                source_node_id=parent_node_id,
                dest_node_id=new_leaf_node.id,
            )

            # print("For letter '{}' added node and edge: {}".format(string[i], new_edge))

            # Add the new leaf node as a child node of the parent node.
            self.add_child_node(parent_node_id, new_leaf_node.id, new_edge.length + 1)

            # Add the string ID to the node collection.
            self.add_string_id(new_leaf_node.id, string_id)
            # print("Added string ID '{}' to new node {}".format(string_id, new_leaf_node))

            # Create suffix link from last parent to current parent (unless it's root).
            self.create_suffix_link(last_parent_node_id, parent_node_id)

            # Remember the last parent node.
            last_parent_node_id = parent_node_id

            # Adjust the active suffix.
            self.follow_suffix_link_or_increment_suffix_first_char(string)

        # Create suffix link from last parent to current parent (unless it's root).
        self.create_suffix_link(last_parent_node_id, parent_node_id)

        # Which is the last leaf node?

        self.suffix.last_char_index += 1
        # print("Incremented suffix last char index: {}".format(self.suffix.last_char_index))

        self._canonize_suffix(self.suffix, string)
        # print("Cannonized suffix: {}".format(self.suffix))

    def _split_edge(self, first_edge, suffix):
        assert isinstance(first_edge, SuffixTreeEdge)
        assert isinstance(suffix, Suffix)

        # Split the label.
        orig_label = first_edge.label
        first_label = orig_label[:suffix.length + 1]
        second_label = orig_label[suffix.length + 1:]

        # print("Splitting edge: {}  <==>  {}".format(first_label[:100], second_label[:100]))

        # Create a new middle node that will split the edge.
        new_middle_node = register_new_node(suffix_node_id=suffix.source_node_id)
        # self._cache_node(new_middle_node)

        # print("New middle node: {}".format(new_middle_node))

        # Create a new edge, from the new middle node to the original destination.
        second_edge_id = make_edge_id(source_node_id=new_middle_node.id, first_char=second_label[0])
        original_dest_node_id = first_edge.dest_node_id

        second_edge = register_new_edge(
            label=second_label,
            edge_id=second_edge_id,
            source_node_id=new_middle_node.id,
            dest_node_id=original_dest_node_id,
        )

        # print("New second edge: {}".format(second_edge))

        # Add the original dest node as a child of the new middle node.
        self.add_child_node(
            parent_node_id=new_middle_node.id,
            child_node_id=original_dest_node_id,
            edge_length=second_edge.length + 1,
        )

        # Shorten the first edge, so its destination is the new middle node.
        try:
            first_edge.shorten(label=first_label, dest_node_id=new_middle_node.id)
            # print("Shortened first edge: {}".format(first_edge))
        except ConcurrencyError as e:
            # If the first edge has changed by now, then abort this split by
            # discarding the new_middle_node, which hasn't been connected to
            # the tree yet, so can safely be discarded.
            new_middle_node.discard()
            sleep(0.1)
            raise e

        # Remove the original dest node as child of the original
        # source node, and add the new middle node instead.
        self.switch_child_node(
            parent_node_id=first_edge.source_node_id,
            old_child_id=original_dest_node_id,
            new_child_id=new_middle_node.id,
            edge_length=first_edge.length + 1,
        )

        # Return middle node.
        return new_middle_node.id

    def add_child_node(self, parent_node_id, child_node_id, edge_length):
        collection = self.get_or_create_node_child_collection(parent_node_id)
        if child_node_id not in collection.child_node_ids:
            # - also pass in the edge length (used when finding strings)
            collection.add_child(child_node_id, edge_length)

    def switch_child_node(self, parent_node_id, old_child_id, new_child_id, edge_length):
        collection = self.get_or_create_node_child_collection(parent_node_id)
        assert isinstance(collection, SuffixTreeNodeChildCollection)
        collection.switch_child(old_child_id, new_child_id, edge_length)

    def add_string_id(self, parent_node_id, string_id):
        collection = self.get_or_create_stringid_collection(parent_node_id)
        if string_id not in collection.items:
            collection.add_item(string_id)

    def create_suffix_link(self, last_parent_node_id, parent_node_id):
        if last_parent_node_id is not None:
            assert parent_node_id is not None

            # Set suffix_node_id on last_parent_node to parent_node_id.
            last_parent_node = self.get_node(last_parent_node_id)
            assert isinstance(last_parent_node, SuffixTreeNode)
            last_parent_node.suffix_node_id = parent_node_id

    def follow_suffix_link_or_increment_suffix_first_char(self, string):
        # If we're at root, there isn't a suffix link, so increment the first char index.
        if self.suffix.source_node_id == self.root_node_id:
            # print("Incrementing suffix first char...")
            self.suffix.first_char_index += 1
        else:
            # If we're not at root, follow the suffix link.
            last_source_node_id = self.suffix.source_node_id
            next_source_node_id = self.get_node(last_source_node_id).suffix_node_id
            # print("Following suffix link from {} to {}".format(last_source_node_id, next_source_node_id))
            self.suffix.source_node_id = next_source_node_id
        self._canonize_suffix(self.suffix, string)

    def is_leaf(self, dest_node):
        return dest_node.is_leaf

    def is_root_node(self, dest_node):
        return dest_node.is_root

    def remove_string(self, string, string_id):
        # Todo: Maybe this is an application service?

        assert isinstance(string_id, six.string_types)
        if self._case_insensitive:
            string = string.lower()
        assert STRING_ID_END not in string_id
        assert STRING_ID_END not in string
        # string += u"{}{}".format(string_id, STRING_ID_END)
        string += u"{}".format(STRING_ID_END)

        # Disassociate the string ID from its leaf nodes.
        #  - but don't discard the node entity, because we might
        #    need to add this suffix back into the tree,
        #    and a hard prune would be relatively complicated.
        #  - although, it might be useful to have something do garbage collection
        from eventsourcing.contrib.suffixtrees.domain.services.generalizedsuffixtree import find_substring_edge

        for i in range(len(string)):

            # Find the suffix in the tree.

            suffix = string[i:]
            edge, ln = find_substring_edge(suffix, self, self._edge_repo)

            if edge is None:
                continue

            # Get the leaf node.
            leaf_node = self.get_node(edge.dest_node_id)
            assert isinstance(leaf_node, SuffixTreeNode)
            if leaf_node.string_id == string_id:
                leaf_node.string_id = None
            try:
                stringid_collection = self._stringid_collection_repo[leaf_node.id]
            except KeyError:
                pass
            else:
                assert isinstance(stringid_collection, StringidCollection)
                if string_id in stringid_collection.items:
                    stringid_collection.remove_item(string_id)

    def get_edge(self, edge_id):
        # Raises KeyError is not found because this method was factored out from various places that
        # used the repo directly and so expected a KeyError to be raised if edge is not in repo.
        return self._edge_repo[edge_id]

    def has_edge(self, edge_id):
        try:
            return self.get_edge(edge_id)
        except KeyError:
            pass

    def get_node(self, node_id):
        # Raises KeyError is not found because this method was factored out from various places that
        # used the repo directly and so expected a KeyError to be raised if node is not in repo.
        # Todo: Change this (and callers) to return (and handle) None when the edge is not found.
        """

        :rtype: SuffixTreeNode
        """
        # if not self._node_repo:
        #     return self.nodes[node_id]
        # try:
        #     return self.nodes[node_id]
        # except KeyError as e:
        #     try:
        #         node = self._node_repo[node_id]
        #     except KeyError:
        #         raise e
        #     else:
        #         self._cache_node(node)
        #         return node

        return self._node_repo[node_id]

    def _cache_node(self, node):
        self.nodes[node.id] = node

    def get_or_create_node_child_collection(self, node_id):
        """
        Returns children of node.

        :rtype: SuffixTreeNodeChildCollection
        """
        try:
            node = self._node_child_collection_repo[node_id]
        except RepositoryKeyError:
            node = register_new_node_child_collection(node_id)
        return node

    def get_or_create_stringid_collection(self, node_id):
        """
        Returns collection of string IDs.

        :rtype: StringidCollection
        """
        try:
            node = self._stringid_collection_repo[node_id]
        except KeyError:
            node = register_new_stringid_collection(node_id)
        return node

    def _canonize_suffix(self, suffix, string):
        """This canonizes the suffix, walking along its suffix string until it
        is explicit or there are no more matched nodes.
        """
        if not suffix.explicit():
            if len(string) < suffix.first_char_index + 1:
                raise AssertionError("String length is {} and suffix first char index is {}"
                                     "".format(len(string), suffix.first_char_index))
            edge_id = make_edge_id(suffix.source_node_id, string[suffix.first_char_index])
            e = self.get_edge(edge_id)
            if e.length <= suffix.length:
                suffix.first_char_index += e.length + 1
                suffix.source_node_id = e.dest_node_id
                self._canonize_suffix(suffix, string)


class SuffixTreeNode(Aggregate):
    """A node in the suffix tree.
    """

    class Created(Created): pass

    class AttributeChanged(AttributeChanged): pass

    class Discarded(Discarded): pass

    def __init__(self, suffix_node_id=None, string_id=None, is_leaf=False, is_root=False, *args, **kwargs):
        super(SuffixTreeNode, self).__init__(*args, **kwargs)
        self._suffix_node_id = suffix_node_id
        self._string_id = string_id
        self._is_leaf = is_leaf
        self._is_root = is_root

    @attribute
    def string_id(self):
        """The id of a string being added to the generalised suffix tree when this node was created.
        """

    @attribute
    def suffix_node_id(self):
        """The ID of the node representing the suffix of the suffix represented by this node.
        """

    @property
    def is_leaf(self):
        return self._is_leaf

    @property
    def is_root(self):
        return self._is_root

    def __repr__(self):
        return "Node(id='{}')".format(self.id)


class StringidCollection(Collection):
    class Created(Collection.Created):
        pass

    class ItemAdded(Collection.ItemAdded):
        pass

    class ItemRemoved(Collection.ItemRemoved):
        pass


class SuffixTreeNodeChildCollection(Aggregate):
    """A collecton of child nodes in the suffix tree.
    """

    class Created(Created): pass

    class AttributeChanged(AttributeChanged): pass

    class Discarded(Discarded): pass

    class ChildNodeAdded(AggregateEvent): pass

    class ChildNodeRemoved(AggregateEvent): pass

    class ChildNodeSwitched(AggregateEvent): pass

    def __init__(self, *args, **kwargs):
        super(SuffixTreeNodeChildCollection, self).__init__(*args, **kwargs)
        self._child_node_ids = OrderedDict()

    @property
    def child_node_ids(self):
        return self._child_node_ids

    def add_child(self, child_node_id, edge_len):
        event = SuffixTreeNodeChildCollection.ChildNodeAdded(
            entity_id=self.id,
            entity_version=self.version,
            child_node_id=child_node_id,
            edge_len=edge_len,
        )
        self._apply(event)
        publish(event)

    def switch_child(self, old_node_id, new_node_id, new_edge_len):
        event = SuffixTreeNodeChildCollection.ChildNodeSwitched(
            entity_id=self.id,
            entity_version=self.version,
            old_node_id=old_node_id,
            new_node_id=new_node_id,
            new_edge_len=new_edge_len,
        )
        self._apply(event)
        publish(event)

    @staticmethod
    def _mutator(event, initial):
        return suffix_tree_node_child_collection_mutator(event, initial)


@singledispatch
def suffix_tree_node_child_collection_mutator(event, initial):
    return entity_mutator(event, initial)


@suffix_tree_node_child_collection_mutator.register(SuffixTreeNodeChildCollection.ChildNodeAdded)
def child_node_child_collection_added_mutator(event, self):
    assert isinstance(self, SuffixTreeNodeChildCollection), self
    self._child_node_ids[event.child_node_id] = event.edge_len
    self._increment_version()
    return self


@suffix_tree_node_child_collection_mutator.register(SuffixTreeNodeChildCollection.ChildNodeSwitched)
def child_node_child_collection_switched_mutator(event, self):
    assert isinstance(self, SuffixTreeNodeChildCollection), self
    try:
        del (self._child_node_ids[event.old_node_id])
    except KeyError:
        pass
    self._child_node_ids[event.new_node_id] = event.new_edge_len
    self._increment_version()
    return self


class SuffixTreeEdge(Aggregate):
    """An edge in the suffix tree.
    """

    class Created(Created):
        pass

    class AttributeChanged(AttributeChanged):
        pass

    class Shortened(AggregateEvent):
        @property
        def label(self):
            return self.__dict__['label']

        @property
        def dest_node_id(self):
            return self.__dict__['dest_node_id']

    class Discarded(Discarded):
        pass

    def __init__(self, label, source_node_id, dest_node_id, **kwargs):
        super(SuffixTreeEdge, self).__init__(**kwargs)
        self._label = label
        self._source_node_id = source_node_id
        self._dest_node_id = dest_node_id

    @attribute
    def label(self):
        """Index of start of string part represented by this edge.
        """
        return self._label

    @attribute
    def source_node_id(self):
        """Id of source node of edge.
        """
        return self._source_node_id

    @attribute
    def dest_node_id(self):
        """Id of destination node of edge.
        """
        return self._dest_node_id

    @property
    def length(self):
        """Number of chars in the string part represented by this edge.
        """
        return len(self.label) - 1

    def __repr__(self):
        return ("Edge(id='{}' source='{}' dest='{}', label='{}')"
                "".format(self.id, self.source_node_id, self.dest_node_id, self.label[:30]))

    def shorten(self, label, dest_node_id):
        self._assert_not_discarded()
        event = SuffixTreeEdge.Shortened(
            entity_id=self._id,
            entity_version=self._version,
            label=label,
            dest_node_id=dest_node_id,
        )
        self._apply(event)
        publish(event)

    @staticmethod
    def _mutator(event, initial):
        return suffix_tree_edge_mutator(event, initial)


@singledispatch
def suffix_tree_edge_mutator(event, initial):
    return entity_mutator(event, initial)


@suffix_tree_edge_mutator.register(SuffixTreeEdge.Shortened)
def suffix_tree_edge_shortened_mutator(event, self):
    assert isinstance(event, SuffixTreeEdge.Shortened), event
    assert isinstance(self, SuffixTreeEdge), type(self)
    self._validate_originator(event)
    self._label = event.label
    self._dest_node_id = event.dest_node_id
    self._increment_version()
    return self


class Suffix(object):
    """Represents a suffix from first_char_index to last_char_index.
    """

    def __init__(self, source_node_id, first_char_index, last_char_index):
        """
        :type source_node_id: six.string_types
            Index of node where this suffix starts.

        :type first_char_index: six.string_types
            Index of start of suffix in string.

        :type last_char_index: six.string_types
            Index of end of suffix in string.
        """
        assert isinstance(source_node_id, six.string_types)
        assert isinstance(first_char_index, six.integer_types)
        assert isinstance(last_char_index, six.integer_types)
        assert source_node_id is not None
        self.source_node_id = source_node_id
        self.first_char_index = first_char_index
        self.last_char_index = last_char_index

    @property
    def length(self):
        """Number of chars in string.
        """
        return self.last_char_index - self.first_char_index

    def explicit(self):
        """A suffix is explicit if it ends on a node. first_char_index
        is set greater than last_char_index to indicate this.
        """
        return self.first_char_index > self.last_char_index

    def implicit(self):
        return self.last_char_index >= self.first_char_index

    def __repr__(self):
        return "Suffix(implicit={} first={}, last={}, node={})".format(self.implicit(), self.first_char_index,
                                                                       self.last_char_index, self.source_node_id)


# Factory methods.

def register_new_node(suffix_node_id=None, string_id=None, is_leaf=False, is_root=False):
    """Factory method, registers new node.
    """
    # Make a unique ID.
    node_id = uuid4().hex

    # Instantiate a Created event.
    created_event = SuffixTreeNode.Created(
        entity_id=node_id,
        suffix_node_id=suffix_node_id,
        string_id=string_id,
        is_leaf=is_leaf,
        is_root=is_root,
    )

    # Make the new Node entity.
    new_node = SuffixTreeNode.mutate(event=created_event)

    # Publish the Created event.
    publish(created_event)

    # Also create a new child collection, to
    # avoid any contention around this event.
    register_new_node_child_collection(node_id)

    # Return the new node.
    return new_node


def register_new_node_child_collection(node_id):
    """Factory method, registers new node child collection.
    """
    event = SuffixTreeNodeChildCollection.Created(
        entity_id=node_id,
    )
    entity = SuffixTreeNodeChildCollection.mutate(event=event)
    publish(event)
    return entity


def register_new_stringid_collection(node_id):
    """Factory method, registers new collection of string IDs.
    """
    event = StringidCollection.Created(
        entity_id=node_id,
    )
    entity = StringidCollection.mutate(event=event)
    publish(event)
    return entity


def make_edge_id(source_node_id, first_char):
    """Returns a string made from given params.
    """
    # assert STRING_ID_PADDING_RIGHT not in first_char
    # assert STRING_ID_PADDING_RIGHT not in source_node_id
    return "{}::{}".format(source_node_id, first_char)


def register_new_edge(edge_id, label, source_node_id, dest_node_id):
    """Factory method, registers new edge.
    """
    event = SuffixTreeEdge.Created(
        entity_id=edge_id,
        label=label,
        source_node_id=source_node_id,
        dest_node_id=dest_node_id,
    )
    entity = SuffixTreeEdge.mutate(event=event)
    publish(event)
    return entity


def register_new_suffix_tree(case_insensitive=False):
    """Factory method, returns new suffix tree object.
    """
    assert isinstance(case_insensitive, bool)

    suffix_tree_id = uuid4().hex

    root_node = register_new_node(is_root=True)
    assert isinstance(root_node, SuffixTreeNode)

    event = GeneralizedSuffixTree.Created(
        entity_id=suffix_tree_id,
        root_node_id=root_node.id,
        case_insensitive=case_insensitive,
    )
    entity = GeneralizedSuffixTree.mutate(event=event)

    assert isinstance(entity, GeneralizedSuffixTree)

    # entity.nodes[root_node.id] = root_node

    publish(event)

    return entity


# Repositories.

class GeneralizedSuffixTreeRepository(AbstractEntityRepository):
    pass


class NodeRepository(AbstractEntityRepository):
    pass


class EdgeRepository(AbstractEntityRepository):
    pass



import datetime
from time import mktime, time
from uuid import UUID, uuid5

import six
from dateutil.relativedelta import relativedelta

from eventsourcing.domain.model.entity import AbstractEntityRepository, Aggregate, AttributeChanged, Created
from eventsourcing.domain.model.events import TimestampedEntityEvent, publish
from eventsourcing.exceptions import RepositoryKeyError
from eventsourcing.utils.time import utc_timezone

Namespace_Timebuckets = UUID('0d7ee297-a976-4c29-91ff-84ffc79d8155')

ONE_YEAR = relativedelta(years=1)
ONE_MONTH = relativedelta(months=1)
ONE_DAY = relativedelta(days=1)
ONE_HOUR = relativedelta(hours=1)
ONE_MINUTE = relativedelta(minutes=1)
ONE_SECOND = relativedelta(seconds=1)

BUCKET_SIZES = {
    'year': ONE_YEAR,
    'month': ONE_MONTH,
    'day': ONE_DAY,
    'hour': ONE_HOUR,
    'minute': ONE_MINUTE,
    'second': ONE_SECOND,
}


class Timebucketedlog(Aggregate):
    class Started(Created):
        pass

    class BucketSizeChanged(AttributeChanged):
        pass

    def __init__(self, name, bucket_size=None, **kwargs):
        super(Timebucketedlog, self).__init__(**kwargs)
        self._name = name
        self._bucket_size = bucket_size

    @property
    def name(self):
        return self._name

    @property
    def started_on(self):
        return self.created_on

    @property
    def bucket_size(self):
        return self._bucket_size

    def append_message(self, message):
        assert isinstance(message, six.string_types)
        bucket_id = make_timebucket_id(self.name, time(), self.bucket_size)
        event = MessageLogged(
            entity_id=bucket_id,
            message=message,
        )
        publish(event)
        return event


class TimebucketedlogRepository(AbstractEntityRepository):
    def get_or_create(self, log_name, bucket_size):
        """
        Gets or creates a log.

        :rtype: Timebucketedlog
        """
        try:
            return self[log_name]
        except RepositoryKeyError:
            return start_new_timebucketedlog(log_name, bucket_size=bucket_size)


def start_new_timebucketedlog(name, bucket_size=None):
    if bucket_size is None:
        bucket_size = 'year'
    if bucket_size not in BUCKET_SIZES:
        raise ValueError("Bucket size '{}' not supported, must be one of: {}"
                         "".format(bucket_size, BUCKET_SIZES.keys()))
    event = Timebucketedlog.Started(
        entity_id=name,
        name=name,
        bucket_size=bucket_size
    )
    entity = Timebucketedlog.mutate(event=event)
    publish(event)
    return entity


class MessageLogged(TimestampedEntityEvent):
    def __init__(self, message, entity_id):
        super(MessageLogged, self).__init__(entity_id=entity_id, message=message)

    @property
    def message(self):
        return self.__dict__['message']

    @property
    def entity_id(self):
        return self.__dict__['entity_id']


def make_timebucket_id(log_id, timestamp, bucket_size):
    d = datetime.datetime.utcfromtimestamp(timestamp)

    assert isinstance(d, datetime.datetime)
    if bucket_size.startswith('year'):
        boundary = '{:04}'.format(
            d.year
        )
    elif bucket_size.startswith('month'):
        boundary = '{:04}-{:02}'.format(
            d.year,
            d.month
        )
    elif bucket_size.startswith('day'):
        boundary = '{:04}-{:02}-{:02}'.format(
            d.year,
            d.month,
            d.day
        )
    elif bucket_size.startswith('hour'):
        boundary = '{:04}-{:02}-{:02}_{:02}'.format(
            d.year,
            d.month,
            d.day,
            d.hour
        )
    elif bucket_size.startswith('minute'):
        boundary = '{:04}-{:02}-{:02}_{:02}-{:02}'.format(
            d.year,
            d.month,
            d.day,
            d.hour,
            d.minute
        )
    elif bucket_size.startswith('second'):
        boundary = '{:04}-{:02}-{:02}_{:02}-{:02}-{:02}'.format(
            d.year,
            d.month,
            d.day,
            d.hour,
            d.minute,
            d.second
        )
    else:
        raise ValueError("Bucket size not supported: {}".format(bucket_size))
    return uuid5(Namespace_Timebuckets, log_id.hex + '_' + boundary)


def next_bucket_starts(timestamp, bucket_size):
    starts = bucket_starts(timestamp, bucket_size)
    duration = bucket_duration(bucket_size)
    return timestamp_from_datetime((starts + duration))


def previous_bucket_starts(timestamp, bucket_size):
    starts = bucket_starts(timestamp, bucket_size)
    duration = bucket_duration(bucket_size)
    return timestamp_from_datetime((starts - duration))


def bucket_starts(timestamp, bucket_size):
    dt = datetime.datetime.utcfromtimestamp(timestamp)
    assert isinstance(dt, datetime.datetime)
    if bucket_size.startswith('year'):
        return datetime.datetime(dt.year, 1, 1, tzinfo=utc_timezone)
    elif bucket_size.startswith('month'):
        return datetime.datetime(dt.year, dt.month, 1, tzinfo=utc_timezone)
    elif bucket_size.startswith('day'):
        return datetime.datetime(dt.year, dt.month, dt.day, tzinfo=utc_timezone)
    elif bucket_size.startswith('hour'):
        return datetime.datetime(dt.year, dt.month, dt.day, dt.hour, tzinfo=utc_timezone)
    elif bucket_size.startswith('minute'):
        return datetime.datetime(dt.year, dt.month, dt.day, dt.hour, dt.minute, tzinfo=utc_timezone)
    elif bucket_size.startswith('second'):
        return datetime.datetime(dt.year, dt.month, dt.day, dt.hour, dt.minute, dt.second, tzinfo=utc_timezone)
    else:
        raise ValueError("Bucket size not supported: {}".format(bucket_size))


def bucket_duration(bucket_size):
    try:
        return BUCKET_SIZES[bucket_size]
    except KeyError:
        raise ValueError("Bucket size not supported: {}. Must be one of: {}"
                         "".format(bucket_size, BUCKET_SIZES.keys()))


# Todo: Move to general utils?
def timestamp_from_datetime(dt):
    assert dt.tzinfo, "Datetime object does not have tzinfo"
    try:
        # Only in Python 3.3+.
        return dt.timestamp()
    except AttributeError:
        return mktime(dt.timetuple()) + (dt.microsecond / 1000000.0)



from abc import ABCMeta

from six import with_metaclass

from eventsourcing.application.policies import CombinedPersistencePolicy
from eventsourcing.infrastructure.activerecord import AbstractActiveRecordStrategy
from eventsourcing.infrastructure.eventstore import EventStore
from eventsourcing.infrastructure.sequenceditemmapper import SequencedItemMapper
from eventsourcing.infrastructure.transcoding import ObjectJSONEncoder, ObjectJSONDecoder


class ReadOnlyEventSourcingApplication(with_metaclass(ABCMeta)):
    def __init__(self, integer_sequenced_active_record_strategy=None,
                 timestamp_sequenced_active_record_strategy=None, always_encrypt=False, cipher=None):
        """
        Constructs an event store using the given stored event repository.

        :param: sequenced_item_repository:  Repository containing sequenced items.

        :param always_encrypt:  Optional encryption of persisted state.

        :param cipher:  Used to encrypt and decrypt stored events.

        """
        assert isinstance(integer_sequenced_active_record_strategy, AbstractActiveRecordStrategy), \
            type(integer_sequenced_active_record_strategy)

        assert isinstance(timestamp_sequenced_active_record_strategy, AbstractActiveRecordStrategy), \
            type(integer_sequenced_active_record_strategy)

        self.integer_sequenced_active_record_strategy = integer_sequenced_active_record_strategy
        self.timestamp_sequenced_active_record_strategy = timestamp_sequenced_active_record_strategy
        self.version_entity_event_store = self.construct_event_store(
            event_sequence_id_attr='entity_id',
            event_position_attr='entity_version',
            active_record_strategy=self.integer_sequenced_active_record_strategy,
            always_encrypt=always_encrypt,
            cipher=cipher,
        )
        self.timestamp_entity_event_store = self.construct_event_store(
            event_sequence_id_attr='entity_id',
            event_position_attr='timestamp',
            active_record_strategy=self.timestamp_sequenced_active_record_strategy,
            always_encrypt=always_encrypt,
            cipher=cipher,
        )

    def construct_event_store(self, event_sequence_id_attr, event_position_attr, active_record_strategy,
                              always_encrypt=False, cipher=None):
        sequenced_item_mapper = self.construct_sequenced_item_mapper(
            sequenced_item_class=active_record_strategy.sequenced_item_class,
            event_sequence_id_attr=event_sequence_id_attr,
            event_position_attr=event_position_attr,
            always_encrypt=always_encrypt,
            cipher=cipher
        )
        event_store = EventStore(
            active_record_strategy=active_record_strategy,
            sequenced_item_mapper=sequenced_item_mapper,
        )
        return event_store

    def construct_sequenced_item_mapper(self, sequenced_item_class, event_sequence_id_attr, event_position_attr,
                                        json_encoder_class=ObjectJSONEncoder, json_decoder_class=ObjectJSONDecoder,
                                        always_encrypt=False, cipher=None):
        return SequencedItemMapper(
            sequenced_item_class=sequenced_item_class,
            event_sequence_id_attr=event_sequence_id_attr,
            event_position_attr=event_position_attr,
            json_encoder_class=json_encoder_class,
            json_decoder_class=json_decoder_class,
            always_encrypt=always_encrypt,
            cipher=cipher
        )

    def close(self):
        self.event_store = None
        self.integer_sequenced_active_record_strategy = None

    def __enter__(self):
        return self

    def __exit__(self, *_):
        self.close()


class EventSourcedApplication(ReadOnlyEventSourcingApplication):
    def __init__(self, **kwargs):
        super(EventSourcedApplication, self).__init__(**kwargs)
        self.persistence_policy = self.construct_persistence_policy()

    def construct_persistence_policy(self):
        return CombinedPersistencePolicy(
            versioned_entity_event_store=self.version_entity_event_store,
            timestamped_entity_event_store=self.timestamp_entity_event_store,
        )

    def close(self):
        if self.persistence_policy is not None:
            self.persistence_policy.close()
            self.persistence_policy = None
        super(EventSourcedApplication, self).close()



from eventsourcing.contrib.suffixtrees.domain.model.generalizedsuffixtree import GeneralizedSuffixTree, \
    GeneralizedSuffixTreeRepository, NodeRepository, SuffixTreeNode, SuffixTreeEdge, EdgeRepository, \
    SuffixTreeNodeChildCollection, StringidCollection
from eventsourcing.infrastructure.eventsourcedrepository import EventSourcedRepository
from eventsourcing.infrastructure.event_sourced_repos.collection_repo import CollectionRepository


class GeneralizedSuffixTreeRepo(EventSourcedRepository, GeneralizedSuffixTreeRepository):
    domain_class = GeneralizedSuffixTree


class NodeRepo(EventSourcedRepository, NodeRepository):
    domain_class = SuffixTreeNode


class NodeChildCollectionRepo(EventSourcedRepository, NodeRepository):
    domain_class = SuffixTreeNodeChildCollection


class EdgeRepo(EventSourcedRepository, EdgeRepository):
    domain_class = SuffixTreeEdge


class StringidCollectionRepo(CollectionRepository):
    domain_class = StringidCollection



from __future__ import absolute_import, division, print_function, unicode_literals

from uuid import uuid4

from eventsourcing.domain.model.entity import Created, Discarded, entity_mutator, singledispatch, Aggregate, \
    AggregateRepository
from eventsourcing.domain.model.events import publish, AggregateEvent


class Collection(Aggregate):
    class Created(Created):
        def __init__(self, **kwargs):
            super(Collection.Created, self).__init__(**kwargs)

    class Discarded(Discarded):
        def __init__(self, **kwargs):
            super(Collection.Discarded, self).__init__(**kwargs)

    class ItemAdded(AggregateEvent):
        @property
        def item(self):
            return self.__dict__['item']

    class ItemRemoved(AggregateEvent):
        @property
        def item(self):
            return self.__dict__['item']

    def __init__(self, **kwargs):
        super(Collection, self).__init__(**kwargs)
        self._items = set()

    def __iter__(self):
        return self._items.__iter__()

    @property
    def items(self):
        self._assert_not_discarded()
        return self._items

    def add_item(self, item):
        self._assert_not_discarded()
        event = self.ItemAdded(
            entity_id=self.id,
            entity_version=self._version,
            item=item,
        )
        self._apply(event)
        publish(event)

    def remove_item(self, item):
        self._assert_not_discarded()
        event = self.ItemRemoved(
            entity_id=self.id,
            entity_version=self._version,
            item=item,
        )
        self._apply(event)
        publish(event)

    @staticmethod
    def _mutator(event, initial):
        return collection_mutator(event, initial)


def register_new_collection(collection_id=None):
    collection_id = uuid4().hex if collection_id is None else collection_id
    event = Collection.Created(entity_id=collection_id)
    entity = Collection.mutate(event=event)
    publish(event)
    return entity


@singledispatch
def collection_mutator(event, initial):
    return entity_mutator(event, initial)


@collection_mutator.register(Collection.ItemAdded)
def collection_item_added_mutator(event, entity):
    assert isinstance(entity, Collection)
    entity._items.add(event.item)
    entity._increment_version()
    return entity


@collection_mutator.register(Collection.ItemRemoved)
def collection_item_removed_mutator(event, entity):
    entity._items.remove(event.item)
    entity._increment_version()
    return entity


class AbstractCollectionRepository(AggregateRepository):
    pass



import os
import sys
from os.path import join, dirname
from subprocess import Popen, PIPE
from unittest.case import TestCase

import eventsourcing


class TestReadmeFile(TestCase):

    def test_code_snippets_in_readme_file(self):
        # Extract lines of Python code from the README.md file.
        readme_filename = 'README.md'
        readme_path = join(dirname(dirname(eventsourcing.__file__)), readme_filename)

        if not os.path.exists(readme_path):
            self.skipTest("Skipped test because usage instructions in README file are "
                          "not available for testing once this package is installed")

        temp_filename = '.README.py'
        temp_path = os.path.join(os.path.dirname(readme_path), temp_filename)
        lines = []
        count_code_lines = 0
        is_code = False
        with open(readme_path) as file:
            for line in file:
                if is_code and line.startswith('```'):
                    is_code = False
                    line = ''
                elif is_code:
                    line = line.strip('\n')
                    count_code_lines += 1
                elif line.startswith('```python'):
                    is_code = True
                    line = ''
                else:
                    line = ''
                lines.append(line)

        self.assertTrue(count_code_lines)

        # Write the code into a temp file.
        with open(temp_path, 'w+') as readme_py:
            readme_py.writelines("\n".join(lines) + '\n')

        # Run the code and catch errors.
        p = Popen([sys.executable, temp_path], stdout=PIPE, stderr=PIPE)
        out, err = p.communicate()
        out = out.decode('utf8').replace(temp_filename, readme_filename)
        err = err.decode('utf8').replace(temp_filename, readme_filename)
        exit_status = p.wait()

        # Check for errors running the code.
        self.assertEqual(exit_status, 0, "Usage exit status {}:\n{}\n{}".format(exit_status, out, err))

        # Delete the temp file.
        os.unlink(temp_path)



from eventsourcing.domain.model.entity import TimestampedVersionedEntity, AbstractEntityRepository, Created
from eventsourcing.domain.model.events import publish, TimestampedVersionedEntityEvent
from eventsourcing.exceptions import RepositoryKeyError


class Sequence(TimestampedVersionedEntity):

    class Started(Created):
        """Occurs when sequence is started."""

    class Appended(TimestampedVersionedEntityEvent):
        """Occurs when item is appended."""

    @property
    def name(self):
        return self.id


def start_sequence(name):
    event = Sequence.Started(entity_id=name)
    entity = Sequence.mutate(event=event)
    publish(event)
    return entity


class SequenceRepository(AbstractEntityRepository):

    def get_or_create(self, sequence_name):
        """
        Gets or creates a log.

        :rtype: Timebucketedlog
        """
        try:
            return self[sequence_name]
        except RepositoryKeyError:
            return start_sequence(sequence_name)



from abc import ABCMeta, abstractmethod

import six


class AbstractCipher(six.with_metaclass(ABCMeta)):

    @abstractmethod
    def encrypt(self, plaintext):
        """Return ciphertext for given plaintext."""

    @abstractmethod
    def decrypt(self, ciphertext):
        """Return plaintext for given ciphertext."""



from collections import namedtuple
from uuid import UUID

from sqlalchemy.sql.schema import Column, Sequence, UniqueConstraint
from sqlalchemy.sql.sqltypes import BigInteger, Float, Integer, String, Text
from sqlalchemy_utils.types.uuid import UUIDType

from eventsourcing.application.policies import PersistencePolicy
from eventsourcing.example.domainmodel import register_new_example
from eventsourcing.example.infrastructure import ExampleRepository
from eventsourcing.infrastructure.eventstore import EventStore
from eventsourcing.infrastructure.sequenceditemmapper import SequencedItemMapper
from eventsourcing.infrastructure.sqlalchemy.activerecords import SQLAlchemyActiveRecordStrategy
from eventsourcing.infrastructure.sqlalchemy.datastore import Base, SQLAlchemyDatastore, SQLAlchemySettings
from eventsourcing.tests.datastore_tests.base import AbstractDatastoreTestCase

# This test replaces the default SequencedItem class with a StoredEvent class.
# How easy is it to customize the infrastructure to support that? We just need
# to define the new sequenced item class, define a suitable active record class,
# and configure the other components. It's easy.


StoredEvent = namedtuple('StoredEvent', ['aggregate_id', 'aggregate_version', 'event_type', 'state'])


class SqlStoredEvent(Base):
    # Explicit table name.
    __tablename__ = 'stored_events'

    # Unique constraint.
    __table_args__ = UniqueConstraint('aggregate_id', 'aggregate_version', name='stored_events_uc'),

    # Primary key.
    id = Column(Integer, Sequence('stored_event_id_seq'), primary_key=True)

    # Sequence ID (e.g. an entity or aggregate ID).
    aggregate_id = Column(UUIDType(), index=True)

    # Position (timestamp) of item in sequence.
    aggregate_version = Column(BigInteger(), index=True)

    # Type of the event (class name).
    event_type = Column(String(100))

    # Timestamp of the event.
    timestamp = Column(Float())

    # State of the item (serialized dict, possibly encrypted).
    state = Column(Text())


class ExampleApplicationWithAlternativeSequencedItemType(object):
    def __init__(self, datastore):
        self.event_store = EventStore(
            active_record_strategy=SQLAlchemyActiveRecordStrategy(
                datastore=datastore,
                active_record_class=SqlStoredEvent,
                sequenced_item_class=StoredEvent,
            ),
            sequenced_item_mapper=SequencedItemMapper(
                sequenced_item_class=StoredEvent,
                event_sequence_id_attr='entity_id',
                event_position_attr='entity_version',
            )
        )
        self.repository = ExampleRepository(
            event_store=self.event_store,
        )
        self.persistence_policy = PersistencePolicy(self.event_store)

    def close(self):
        self.persistence_policy.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class TestExampleWithAlternativeSequencedItemType(AbstractDatastoreTestCase):
    def setUp(self):
        super(TestExampleWithAlternativeSequencedItemType, self).setUp()
        self.datastore.setup_connection()
        self.datastore.setup_tables()

    def tearDown(self):
        self.datastore.drop_tables()
        self.datastore.drop_connection()
        super(TestExampleWithAlternativeSequencedItemType, self).setUp()

    def construct_datastore(self):
        return SQLAlchemyDatastore(
            base=Base,
            settings=SQLAlchemySettings(),
            tables=(SqlStoredEvent,),
        )

    def test(self):
        with ExampleApplicationWithAlternativeSequencedItemType(self.datastore) as app:
            # Create entity.
            entity1 = register_new_example(a='a', b='b')
            self.assertIsInstance(entity1.id, UUID)
            self.assertEqual(entity1.a, 'a')
            self.assertEqual(entity1.b, 'b')

            # Check there is a stored event.
            all_records = list(app.event_store.active_record_strategy.filter())
            assert len(all_records) == 1
            stored_event = all_records[0]
            assert stored_event.aggregate_id == entity1.id
            assert stored_event.aggregate_version == 0

            # Todo: Finish this off so we can read events.
            # Read entity from repo.
            retrieved_obj = app.repository[entity1.id]
            self.assertEqual(retrieved_obj.id, entity1.id)



from eventsourcing.tests.example_application_tests.base import ExampleApplicationTestCase
from eventsourcing.tests.sequenced_item_tests.test_sqlalchemy_active_record_strategy import \
    WithSQLAlchemyActiveRecordStrategies


class TestExampleApplicationWithSQLAlchemy(WithSQLAlchemyActiveRecordStrategies, ExampleApplicationTestCase):
    pass



from abc import abstractproperty

from eventsourcing.domain.model.entity import AbstractEntityRepository
from eventsourcing.exceptions import RepositoryKeyError
from eventsourcing.infrastructure.eventplayer import EventPlayer
from eventsourcing.infrastructure.eventstore import AbstractEventStore
from eventsourcing.infrastructure.snapshotting import entity_from_snapshot


class EventSourcedRepository(AbstractEntityRepository):

    # If the entity won't have very many events, marking the entity as
    # "short" by setting __is_short__ value equal to True will mean
    # the fastest path for getting all the events is used. If you set
    # a value for page size (see below), this option will have no effect.
    __is_short__ = False

    # The page size by which events are retrieved. If this
    # value is set to a positive integer, the events of
    # the entity will be retrieved in pages, using a series
    # of queries, rather than with one potentially large query.
    __page_size__ = None

    # The mutator function used by this repository. Can either
    # be set as a class attribute, or passed as a constructor arg.
    mutator = None

    def __init__(self, event_store, mutator=None, snapshot_strategy=None, use_cache=False):
        self._cache = {}
        self._snapshot_strategy = snapshot_strategy
        # self._use_cache = use_cache

        # Check we got an event store.
        assert isinstance(event_store, AbstractEventStore)
        self.event_store = event_store

        # Instantiate an event player for this repo.
        mutator = mutator or type(self).mutator
        if mutator is None:
            raise ValueError("Repository needs a mutator function (set class attribute or pass constructor arg)")
        self.event_player = EventPlayer(
            event_store=self.event_store,
            mutator=mutator,
            page_size=self.__page_size__,
            is_short=self.__is_short__,
            snapshot_strategy=self._snapshot_strategy,
        )

    def __contains__(self, entity_id):
        """
        Returns a boolean value according to whether entity with given ID exists.
        """
        return self.get_entity(entity_id) is not None

    def __getitem__(self, entity_id):
        """
        Returns entity with given ID.
        """
        # # Get entity from the cache.
        # if self._use_cache:
        #     try:
        #         return self._cache[entity_id]
        #     except KeyError:
        #         pass

        # Reconstitute the entity.
        entity = self.get_entity(entity_id)

        # Never created or already discarded?
        if entity is None:
            raise RepositoryKeyError(entity_id)

        # # Put entity in the cache.
        # if self._use_cache:
        #     self.add_cache(entity_id, entity)

        # Return entity.
        return entity

    # def add_cache(self, entity_id, entity):
    #     self._cache[entity_id] = entity

    def get_entity(self, entity_id, lte=None):
        """
        Returns entity with given ID, optionally as it was until the given domain event ID.
        """

        # Get a snapshot (None if none exist).
        if self._snapshot_strategy is not None:
            snapshot = self._snapshot_strategy.get_snapshot(entity_id, lte=lte)
        else:
            snapshot = None

        # Decide the initial state, and after when we need to get the events.
        if snapshot is None:
            initial_state = None
            gte = None
        else:
            initial_state = entity_from_snapshot(snapshot)
            gte = initial_state._version

        # Replay domain events.
        return self.event_player.replay_entity(entity_id, gte=gte, lte=lte, initial_state=initial_state)

    # def fastforward(self, stale_entity, lt=None, lte=None):
    #     """
    #     Mutates an instance of an entity, according to the events that have occurred since its version.
    #     """
    #     return self.event_player.fastforward(stale_entity, lt=lt, lte=lte)



import base64
import zlib

from Crypto import Random
from Crypto.Cipher import AES

from eventsourcing.domain.services.cipher import AbstractCipher


class AESCipher(AbstractCipher):

    BLOCK_SIZE = 16

    def __init__(self, aes_key):
        self.aes_key = aes_key

        # Pick AES mode.
        self.aes_mode = AES.MODE_CBC

        # Fix the block size.
        self.bs = AES.block_size

    def encrypt(self, plaintext):
        # Create a unique initialisation vector each time something is encrypted.
        iv = Random.new().read(self.bs)

        # Create an AES cipher.
        cipher = AES.new(self.aes_key, self.aes_mode, iv)

        # Construct the ciphertext string.
        ciphertext = base64.b64encode(
            iv + cipher.encrypt(
                self._pad(
                    base64.b64encode(
                        zlib.compress(
                            plaintext.encode('utf8')
                        )
                    )
                )
            )
        ).decode('utf8')

        return ciphertext

    def decrypt(self, ciphertext):
        # Recover the initialisation vector.
        ciphertext_bytes_base64 = ciphertext.encode('utf8')
        ciphertext_bytes = base64.b64decode(ciphertext_bytes_base64)
        iv = ciphertext_bytes[:self.bs]

        # Create the AES cipher.
        cipher = AES.new(self.aes_key, self.aes_mode, iv)

        # Construct the plaintext string.
        plaintext = zlib.decompress(
            base64.b64decode(
                self._unpad(
                    cipher.decrypt(
                        ciphertext_bytes[self.bs:]
                    )
                )
            )
        ).decode('utf8')

        return plaintext

    def _pad(self, s):
        padding_size = self.bs - len(s) % self.bs
        return s + padding_size * chr(padding_size).encode('utf8')

    @staticmethod
    def _unpad(s):
        return s[:-ord(s[len(s)-1:])]



import six

from eventsourcing.contrib.suffixtrees.domain.model.generalizedsuffixtree import SuffixTreeNode, STRING_ID_END, \
    GeneralizedSuffixTree, EdgeRepository, make_edge_id, StringidCollection
from eventsourcing.exceptions import RepositoryKeyError


def get_string_ids(node_id, node_repo, node_child_collection_repo, stringid_collection_repo, length_until_end=0, edge_length=0,
                   limit=None, hop_count=0, hop_max=None):
    """Generator that yields unique string IDs from leaf nodes as they are discovered,
    by performing a depth first search on the suffix tree from the given node.
    """
    stack = list()

    # _print("Searching for string IDs from node: {}".format(node_id))

    stack.append((node_id, edge_length, None))
    unique_node_ids = set(node_id)
    unique_string_ids = set()
    cum_lengths_until_end = {None: length_until_end}
    while stack:

        (node_id, edge_length, parent_node_id) = stack.pop()

        length_until_end = cum_lengths_until_end[parent_node_id] + edge_length
        cum_lengths_until_end[node_id] = length_until_end

        node = node_repo[node_id]
        assert isinstance(node, SuffixTreeNode)

        try:
            stringid_collection = stringid_collection_repo[node_id]
        except KeyError:
            stringid_collection = None
        else:
            assert isinstance(stringid_collection, StringidCollection)

        if stringid_collection and stringid_collection.items:

            # Deduplicate string IDs (the substring might match more than one suffix in any given string).
            for string_id in stringid_collection.items:
                if string_id in unique_string_ids:
                    continue

                # Check the match doesn't encroach upon the string's extension.
                extension_length = len(STRING_ID_END)
                if length_until_end < extension_length:
                    continue

                # print("Found string ID '{}' at node: {}".format(string_id, node_id))

                # Remember the string ID, we only want one node per string ID.
                unique_string_ids.add(string_id)

                # Yield the string ID.
                yield string_id

                # Check the limit, stop if we've had enough string IDs.
                if limit is not None and len(unique_string_ids) >= limit:
                    raise StopIteration


        # Check the hop count before getting the child collection.
        if hop_max is None or hop_count < hop_max:
            try:
                node_child_collection = node_child_collection_repo[node_id]
            except RepositoryKeyError:
                # It doesn't matter if there isn't a child collection.
                pass
            else:
                # Since there's a collection, put the child nodes on the stack.
                child_node_ids = node_child_collection._child_node_ids.copy()
                for (child_node_id, edge_length) in child_node_ids.items():
                    # Check the hop count, stop if we've seen enough tree nodes.
                    #  - this works by choking off the stack, but we continue
                    #    to process everything on the stack after this condition
                    #    which saves putting loads of things on the stack that
                    #    won't be used, and walking more tree than is necessary
                    if hop_max is not None and hop_count >= hop_max:
                        break

                    # Put the child node on the stack.
                    # if child_node_id not in unique_node_ids:
                    if child_node_id != node_id:
                        unique_node_ids.add(child_node_id)
                        stack.append((child_node_id, edge_length, node.id))

                    # Increase the hop count.
                    hop_count += 1


def find_substring_edge(substring, suffix_tree, edge_repo):
    """Returns the last edge, if substring in tree, otherwise None.
    """
    assert isinstance(substring, six.string_types)
    assert isinstance(suffix_tree, GeneralizedSuffixTree)
    assert isinstance(edge_repo, EdgeRepository)
    if not substring:
        return None, None
    if suffix_tree.case_insensitive:
        substring = substring.lower()
    curr_node_id = suffix_tree.root_node_id
    i = 0
    while i < len(substring):
        edge_id = make_edge_id(curr_node_id, substring[i])
        try:
            edge = edge_repo[edge_id]
        except RepositoryKeyError:
            return None, None
        ln = min(edge.length + 1, len(substring) - i)
        if substring[i:i + ln] != edge.label[:ln]:
            return None, None
        i += edge.length + 1
        curr_node_id = edge.dest_node_id
    return edge, ln


def has_substring(substring, suffix_tree, edge_repo):
    edge, ln = find_substring_edge(substring, suffix_tree, edge_repo)
    return edge is not None



import six
from sqlalchemy.exc import IntegrityError
from sqlalchemy.sql.expression import asc, desc
from sqlalchemy.sql.schema import Column, Sequence, UniqueConstraint
from sqlalchemy.sql.sqltypes import BigInteger, Float, Integer, String, Text
from sqlalchemy_utils.types.uuid import UUIDType

from eventsourcing.infrastructure.activerecord import AbstractActiveRecordStrategy
from eventsourcing.infrastructure.sqlalchemy.datastore import Base, SQLAlchemyDatastore


class SQLAlchemyActiveRecordStrategy(AbstractActiveRecordStrategy):

    def __init__(self, datastore, *args, **kwargs):
        assert isinstance(datastore, SQLAlchemyDatastore)
        super(SQLAlchemyActiveRecordStrategy, self).__init__(*args, **kwargs)
        self.datastore = datastore

    def append_item(self, sequenced_item):
        active_record = self.to_active_record(sequenced_item)
        try:
            # Write stored event into the transaction.
            self.add_record_to_session(active_record)

            # Commit the transaction.
            self.datastore.db_session.commit()

        except IntegrityError as e:
            # Roll back the transaction.
            self.datastore.db_session.rollback()
            self.raise_sequenced_item_error(sequenced_item, e)
        finally:
            # Begin new transaction.
            self.datastore.db_session.close()

    def get_item(self, sequence_id, eq):
        try:
            filter_args = {self.field_names.sequence_id: sequence_id}
            query = self.filter(**filter_args)
            position_field = getattr(self.active_record_class, self.field_names.position)
            query = query.filter(position_field == eq)
            events = six.moves.map(self.from_active_record, query)
            events = list(events)
        finally:
            self.datastore.db_session.close()

        try:
            return events[0]
        except IndexError:
            self.raise_index_error(eq)

    def get_items(self, sequence_id, gt=None, gte=None, lt=None, lte=None, limit=None,
                  query_ascending=True, results_ascending=True):

        assert limit is None or limit >= 1, limit

        try:
            filter_args = {self.field_names.sequence_id: sequence_id}
            query = self.filter(**filter_args)

            position_field = getattr(self.active_record_class, self.field_names.position)

            if query_ascending:
                query = query.order_by(asc(position_field))
            else:
                query = query.order_by(desc(position_field))

            if gt is not None:
                query = query.filter(position_field > gt)
            if gte is not None:
                query = query.filter(position_field >= gte)
            if lt is not None:
                query = query.filter(position_field < lt)
            if lte is not None:
                query = query.filter(position_field <= lte)

            if limit is not None:
                query = query.limit(limit)

            events = six.moves.map(self.from_active_record, query)
            events = list(events)

        finally:
            self.datastore.db_session.close()

        if results_ascending != query_ascending:
            events.reverse()

        return events

    def all_items(self):
        return map(self.from_active_record, self.filter())

    def add_record_to_session(self, active_record):
        if isinstance(active_record, list):
            [self.add_record_to_session(r) for r in active_record]
        else:
            self.datastore.db_session.add(active_record)

    def to_active_record(self, sequenced_item):
        """
        Returns an active record, from given sequenced item.
        """
        # Recurse if it's a list.
        if isinstance(sequenced_item, list):
            return [self.to_active_record(i) for i in sequenced_item]

        # Check we got a sequenced item.
        assert isinstance(sequenced_item, self.sequenced_item_class), (self.sequenced_item_class, type(sequenced_item))

        # Construct and return an ORM object.
        orm_kwargs = {f: sequenced_item[i] for i, f in enumerate(self.field_names)}
        return self.active_record_class(**orm_kwargs)

    def from_active_record(self, active_record):
        """
        Returns a sequenced item, from given active record.
        """
        item_args = [getattr(active_record, f) for f in self.field_names]
        return self.sequenced_item_class(*item_args)

    def filter(self, *args, **kwargs):
        query = self.datastore.db_session.query(self.active_record_class)
        return query.filter_by(*args, **kwargs)


class SqlIntegerSequencedItem(Base):
    __tablename__ = 'integer_sequenced_items'

    id = Column(Integer, Sequence('integer_sequened_item_id_seq'), primary_key=True)

    # Sequence ID (e.g. an entity or aggregate ID).
    sequence_id = Column(UUIDType(), index=True)

    # Position (index) of item in sequence.
    position = Column(BigInteger(), index=True)

    # Topic of the item (e.g. path to domain event class).
    topic = Column(String(255))

    # State of the item (serialized dict, possibly encrypted).
    data = Column(Text())

    # Unique constraint includes 'entity_id' which is a good value
    # to partition on, because all events for an entity will be in the same
    # partition, which may help performance.
    __table_args__ = UniqueConstraint('sequence_id', 'position',
                                      name='integer_sequenced_item_uc'),


class SqlTimestampSequencedItem(Base):

    # Explicit table name.
    __tablename__ = 'timestamp_sequenced_items'

    # Unique constraint.
    __table_args__ = UniqueConstraint('sequence_id', 'position', name='time_sequenced_items_uc'),

    # Primary key.
    id = Column(Integer, Sequence('integer_sequened_item_id_seq'), primary_key=True)

    # Sequence ID (e.g. an entity or aggregate ID).
    sequence_id = Column(UUIDType(), index=True)

    # Position (timestamp) of item in sequence.
    position = Column(Float(), index=True)

    # Topic of the item (e.g. path to domain event class).
    topic = Column(String(255))

    # State of the item (serialized dict, possibly encrypted).
    data = Column(Text())



import os

from sqlalchemy import create_engine
from sqlalchemy.ext.declarative.api import declarative_base
from sqlalchemy.orm import scoped_session, sessionmaker

from eventsourcing.infrastructure.datastore import Datastore, DatastoreConnectionError, DatastoreSettings

Base = declarative_base()


DEFAULT_SQLALCHEMY_DB_URI = 'sqlite:///:memory:'


class SQLAlchemySettings(DatastoreSettings):
    DB_URI = os.getenv('DB_URI', DEFAULT_SQLALCHEMY_DB_URI)

    def __init__(self, uri=DEFAULT_SQLALCHEMY_DB_URI):
        self.uri = uri or self.DB_URI


class SQLAlchemyDatastore(Datastore):

    def __init__(self, base, **kwargs):
        super(SQLAlchemyDatastore, self).__init__(**kwargs)
        self._db_session = None
        self._engine = None
        self._base = base

    def setup_connection(self):
        assert isinstance(self.settings, SQLAlchemySettings), self.settings
        self._engine = create_engine(self.settings.uri, strategy='threadlocal')

    def setup_tables(self):
        self._base.metadata.create_all(self._engine)

    def drop_connection(self):
        if self._db_session:
            self._db_session.close()
        if self._engine:
            self._engine = None

    def drop_tables(self):
        self._base.metadata.drop_all(self._engine)

    @property
    def db_session(self):
        if self._engine is None:
            raise DatastoreConnectionError("Need to call setup_connection() first")
        if self._db_session is None:
            session_factory = sessionmaker(bind=self._engine)
            self._db_session = scoped_session(session_factory)
        return self._db_session



from abc import ABCMeta, abstractmethod

import six

from eventsourcing.exceptions import SequencedItemError
from eventsourcing.infrastructure.sequenceditem import SequencedItemFieldNames


class AbstractActiveRecordStrategy(six.with_metaclass(ABCMeta)):

    def __init__(self, active_record_class, sequenced_item_class):
        self.active_record_class = active_record_class
        self.sequenced_item_class = sequenced_item_class
        self.field_names = SequencedItemFieldNames(self.sequenced_item_class)

    @abstractmethod
    def append_item(self, sequenced_item):
        """
        Writes sequenced item into the datastore.
        """

    @abstractmethod
    def get_item(self, sequence_id, eq):
        """
        Reads sequenced item from the datastore.
        """

    @abstractmethod
    def get_items(self, sequence_id, gt=None, gte=None, lt=None, lte=None, limit=None,
                  query_ascending=True, results_ascending=True):
        """
        Reads sequenced items from the datastore.
        """

    @abstractmethod
    def all_items(self):
        """
        Returns all items from all sequences (possibly in chronological order, depending on database).
        """

    def raise_sequenced_item_error(self, sequenced_item, e):
        sequenced_item = sequenced_item[0] if isinstance(sequenced_item, list) else sequenced_item
        raise SequencedItemError("Item at position '{}' already exists in sequence '{}': {}"
                                 "".format(sequenced_item[1], sequenced_item[0], e))

    def raise_index_error(self, eq):
        raise IndexError("Sequence index out of range: {}".format(eq))



from eventsourcing.domain.model.events import TimestampedEntityEvent, VersionedEntityEvent, subscribe, unsubscribe
from eventsourcing.infrastructure.eventstore import AbstractEventStore


class PersistencePolicy(object):
    """
    Stores events of given type to given event store, whenever they are published.
    """
    def __init__(self, event_store, event_type=None):
        assert isinstance(event_store, AbstractEventStore)
        self.event_store = event_store
        self.event_type = event_type
        subscribe(self.store_event, self.is_event)

    def is_event(self, event):
        if isinstance(event, (list, tuple)):
            return all(map(self.is_event, event))
        return self.event_type is None or isinstance(event, self.event_type)

    def store_event(self, event):
        self.event_store.append(event)

    def close(self):
        unsubscribe(self.store_event, self.is_event)


class CombinedPersistencePolicy(object):
    """
    Persists both timestamped and versioned entity events, whenever they are published.
    """

    def __init__(self, timestamped_entity_event_store, versioned_entity_event_store):
        self.timestamped_entity_event_policy = PersistencePolicy(
            event_store=timestamped_entity_event_store,
            event_type=TimestampedEntityEvent,
        )
        self.versioned_entity_event_policy = PersistencePolicy(
            event_store=versioned_entity_event_store,
            event_type = VersionedEntityEvent,
        )

    def close(self):
        self.timestamped_entity_event_policy.close()
        self.versioned_entity_event_policy.close()



import os
import unittest
from unittest import TestCase


class AbstractTestCase(TestCase):
    """
    Base class for test cases with abstract test_* methods.
    """

    def setUp(self):
        """
        Returns None if test case class ends with 'TestCase',
        which means the test case isn't included in the suite.
        """
        super(AbstractTestCase, self).setUp()
        if type(self).__name__.endswith('TestCase'):
            self.skipTest('Ignored abstract test.')
        else:
            super(AbstractTestCase, self).setUp()


def notquick(*args, **kwargs):
    return unittest.skipIf(os.getenv("QUICK_TESTS_ONLY"), 'Ignored slow test.')



from eventsourcing.contrib.suffixtrees.domain.model.suffixtree import Edge, EdgeRepository, Node, NodeRepository, \
    SuffixTree, SuffixTreeRepository
from eventsourcing.infrastructure.eventsourcedrepository import EventSourcedRepository


class SuffixTreeRepo(EventSourcedRepository, SuffixTreeRepository):
    domain_class = SuffixTree


class NodeRepo(EventSourcedRepository, NodeRepository):
    domain_class = Node


class EdgeRepo(EventSourcedRepository, EdgeRepository):
    domain_class = Edge
